# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0.

# This file is generated

from . import model
from .client import GreengrassCoreIPCClient
from . import client

import concurrent.futures
import datetime
import typing


class GreengrassCoreIPCClientV2:
    """
    V2 client for the GreengrassCoreIPC service.

    Args:
        client: Connection that this client will use. If you do not provide one, it will be made automatically.
        executor: Executor used to run on_stream_event and on_stream_closed callbacks to avoid blocking the networking
         thread. By default, a ThreadPoolExecutor will be created and used. Use None to run callbacks in the
         networking thread, but understand that your code can deadlock the networking thread if it performs a
         synchronous network call.
    """

    def __init__(self, client: typing.Optional[GreengrassCoreIPCClient] = None,
                 executor: typing.Optional[concurrent.futures.Executor] = True):
        if client is None:
            import awsiot.greengrasscoreipc
            client = awsiot.greengrasscoreipc.connect()
        self.client = client
        if executor is True:
            executor = concurrent.futures.ThreadPoolExecutor()
        self.executor = executor
        self.ignore_executor_exceptions = False

    def close(self, *, executor_wait=True) -> concurrent.futures.Future:
        """
        Close the underlying connection and shutdown the event executor (if any)

        Args:
            executor_wait: If true (default), then this method will block until the executor finishes running
                all tasks and shuts down.

        Returns:
            The future which will complete
            when the shutdown process is done. The future will have an
            exception if shutdown was caused by an error, or a result
            of None if the shutdown was clean and user-initiated.
        """
        fut = self.client.close()

        # events that arrive during the shutdown process will generate executor exceptions, ignore them
        self.ignore_executor_exceptions	= True
        if self.executor is not None:
            self.executor.shutdown(wait=executor_wait)
        return fut

    def __combine_futures(self, future1: concurrent.futures.Future,
                          future2: concurrent.futures.Future) -> concurrent.futures.Future:
        def callback(*args, **kwargs):
            try:
                future1.result()
            except Exception as e:
                future2.set_exception(e)
        future1.add_done_callback(callback)
        return future2

    @staticmethod
    def __handle_error():
        import sys
        import traceback
        traceback.print_exc(file=sys.stderr)

    def __wrap_error(self, func):
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                self.__handle_error()
                raise e
        return wrapper

    def __create_stream_handler(real_self, operation, on_stream_event, on_stream_error, on_stream_closed):
        stream_handler_type = type(operation + 'Handler', (getattr(client, operation + "StreamHandler"),), {})
        if on_stream_event is not None:
            on_stream_event = real_self.__wrap_error(on_stream_event)
            def handler(self, event):
                if real_self.executor is not None:
                    try:
                        real_self.executor.submit(on_stream_event, event)
                    except RuntimeError:
                        if not real_self.ignore_executor_exceptions:
                            raise
                else:
                    on_stream_event(event)
            setattr(stream_handler_type, "on_stream_event", handler)
        if on_stream_error is not None:
            on_stream_error = real_self.__wrap_error(on_stream_error)
            def handler(self, error):
                return on_stream_error(error)
            setattr(stream_handler_type, "on_stream_error", handler)
        if on_stream_closed is not None:
            on_stream_closed = real_self.__wrap_error(on_stream_closed)
            def handler(self):
                if real_self.executor is not None:
                    try:
                        real_self.executor.submit(on_stream_closed)
                    except RuntimeError:
                        if real_self.ignore_executor_exceptions:
                            raise
                else:
                    on_stream_closed()
            setattr(stream_handler_type, "on_stream_closed", handler)
        return stream_handler_type()

    def __handle_stream_handler(real_self, operation, stream_handler, on_stream_event, on_stream_error, on_stream_closed):
        if stream_handler is not None and (on_stream_event is not None or on_stream_error is not None or on_stream_closed is not None):
            raise ValueError("Must choose either stream_handler or on_stream_event/on_stream_error/on_stream_closed")
        if stream_handler is not None and real_self.executor is not None:
            return real_self.__create_stream_handler(operation, stream_handler.on_stream_event,
                                                     stream_handler.on_stream_error, stream_handler.on_stream_closed)
        if stream_handler is None:
            return real_self.__create_stream_handler(operation, on_stream_event, on_stream_error, on_stream_closed)
        return stream_handler

    def authorize_client_device_action(self, *,
        client_device_auth_token: typing.Optional[str] = None,
        operation: typing.Optional[str] = None,
        resource: typing.Optional[str] = None) -> model.AuthorizeClientDeviceActionResponse:
        """
        Perform the AuthorizeClientDeviceAction operation synchronously.

        Args:
            client_device_auth_token: 
            operation: 
            resource: 
        """
        return self.authorize_client_device_action_async(client_device_auth_token=client_device_auth_token, operation=operation, resource=resource).result()

    def authorize_client_device_action_async(self, *,
        client_device_auth_token: typing.Optional[str] = None,
        operation: typing.Optional[str] = None,
        resource: typing.Optional[str] = None):  # type: (...) -> concurrent.futures.Future[model.AuthorizeClientDeviceActionResponse]
        """
        Perform the AuthorizeClientDeviceAction operation asynchronously.

        Args:
            client_device_auth_token: 
            operation: 
            resource: 
        """
        request = model.AuthorizeClientDeviceActionRequest(client_device_auth_token=client_device_auth_token, operation=operation, resource=resource)
        operation = self.client.new_authorize_client_device_action()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def cancel_local_deployment(self, *,
        deployment_id: typing.Optional[str] = None) -> model.CancelLocalDeploymentResponse:
        """
        Perform the CancelLocalDeployment operation synchronously.

        Args:
            deployment_id: 
        """
        return self.cancel_local_deployment_async(deployment_id=deployment_id).result()

    def cancel_local_deployment_async(self, *,
        deployment_id: typing.Optional[str] = None):  # type: (...) -> concurrent.futures.Future[model.CancelLocalDeploymentResponse]
        """
        Perform the CancelLocalDeployment operation asynchronously.

        Args:
            deployment_id: 
        """
        request = model.CancelLocalDeploymentRequest(deployment_id=deployment_id)
        operation = self.client.new_cancel_local_deployment()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def create_debug_password(self) -> model.CreateDebugPasswordResponse:
        """
        Perform the CreateDebugPassword operation synchronously.

        """
        return self.create_debug_password_async().result()

    def create_debug_password_async(self):  # type: (...) -> concurrent.futures.Future[model.CreateDebugPasswordResponse]
        """
        Perform the CreateDebugPassword operation asynchronously.

        """
        request = model.CreateDebugPasswordRequest()
        operation = self.client.new_create_debug_password()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def create_local_deployment(self, *,
        group_name: typing.Optional[str] = None,
        root_component_versions_to_add: typing.Optional[typing.Dict[str, str]] = None,
        root_components_to_remove: typing.Optional[typing.List[str]] = None,
        component_to_configuration: typing.Optional[typing.Dict[str, typing.Dict[str, typing.Any]]] = None,
        component_to_run_with_info: typing.Optional[typing.Dict[str, model.RunWithInfo]] = None,
        recipe_directory_path: typing.Optional[str] = None,
        artifacts_directory_path: typing.Optional[str] = None,
        failure_handling_policy: typing.Optional[str] = None) -> model.CreateLocalDeploymentResponse:
        """
        Perform the CreateLocalDeployment operation synchronously.

        Args:
            group_name: 
            root_component_versions_to_add: 
            root_components_to_remove: 
            component_to_configuration: 
            component_to_run_with_info: 
            recipe_directory_path: 
            artifacts_directory_path: 
            failure_handling_policy: FailureHandlingPolicy enum value
        """
        return self.create_local_deployment_async(group_name=group_name, root_component_versions_to_add=root_component_versions_to_add, root_components_to_remove=root_components_to_remove, component_to_configuration=component_to_configuration, component_to_run_with_info=component_to_run_with_info, recipe_directory_path=recipe_directory_path, artifacts_directory_path=artifacts_directory_path, failure_handling_policy=failure_handling_policy).result()

    def create_local_deployment_async(self, *,
        group_name: typing.Optional[str] = None,
        root_component_versions_to_add: typing.Optional[typing.Dict[str, str]] = None,
        root_components_to_remove: typing.Optional[typing.List[str]] = None,
        component_to_configuration: typing.Optional[typing.Dict[str, typing.Dict[str, typing.Any]]] = None,
        component_to_run_with_info: typing.Optional[typing.Dict[str, model.RunWithInfo]] = None,
        recipe_directory_path: typing.Optional[str] = None,
        artifacts_directory_path: typing.Optional[str] = None,
        failure_handling_policy: typing.Optional[str] = None):  # type: (...) -> concurrent.futures.Future[model.CreateLocalDeploymentResponse]
        """
        Perform the CreateLocalDeployment operation asynchronously.

        Args:
            group_name: 
            root_component_versions_to_add: 
            root_components_to_remove: 
            component_to_configuration: 
            component_to_run_with_info: 
            recipe_directory_path: 
            artifacts_directory_path: 
            failure_handling_policy: FailureHandlingPolicy enum value
        """
        request = model.CreateLocalDeploymentRequest(group_name=group_name, root_component_versions_to_add=root_component_versions_to_add, root_components_to_remove=root_components_to_remove, component_to_configuration=component_to_configuration, component_to_run_with_info=component_to_run_with_info, recipe_directory_path=recipe_directory_path, artifacts_directory_path=artifacts_directory_path, failure_handling_policy=failure_handling_policy)
        operation = self.client.new_create_local_deployment()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def defer_component_update(self, *,
        deployment_id: typing.Optional[str] = None,
        message: typing.Optional[str] = None,
        recheck_after_ms: typing.Optional[int] = None) -> model.DeferComponentUpdateResponse:
        """
        Perform the DeferComponentUpdate operation synchronously.

        Args:
            deployment_id: 
            message: 
            recheck_after_ms: 
        """
        return self.defer_component_update_async(deployment_id=deployment_id, message=message, recheck_after_ms=recheck_after_ms).result()

    def defer_component_update_async(self, *,
        deployment_id: typing.Optional[str] = None,
        message: typing.Optional[str] = None,
        recheck_after_ms: typing.Optional[int] = None):  # type: (...) -> concurrent.futures.Future[model.DeferComponentUpdateResponse]
        """
        Perform the DeferComponentUpdate operation asynchronously.

        Args:
            deployment_id: 
            message: 
            recheck_after_ms: 
        """
        request = model.DeferComponentUpdateRequest(deployment_id=deployment_id, message=message, recheck_after_ms=recheck_after_ms)
        operation = self.client.new_defer_component_update()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def delete_thing_shadow(self, *,
        thing_name: typing.Optional[str] = None,
        shadow_name: typing.Optional[str] = None) -> model.DeleteThingShadowResponse:
        """
        Perform the DeleteThingShadow operation synchronously.

        Args:
            thing_name: 
            shadow_name: 
        """
        return self.delete_thing_shadow_async(thing_name=thing_name, shadow_name=shadow_name).result()

    def delete_thing_shadow_async(self, *,
        thing_name: typing.Optional[str] = None,
        shadow_name: typing.Optional[str] = None):  # type: (...) -> concurrent.futures.Future[model.DeleteThingShadowResponse]
        """
        Perform the DeleteThingShadow operation asynchronously.

        Args:
            thing_name: 
            shadow_name: 
        """
        request = model.DeleteThingShadowRequest(thing_name=thing_name, shadow_name=shadow_name)
        operation = self.client.new_delete_thing_shadow()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def get_client_device_auth_token(self, *,
        credential: typing.Optional[model.CredentialDocument] = None) -> model.GetClientDeviceAuthTokenResponse:
        """
        Perform the GetClientDeviceAuthToken operation synchronously.

        Args:
            credential: 
        """
        return self.get_client_device_auth_token_async(credential=credential).result()

    def get_client_device_auth_token_async(self, *,
        credential: typing.Optional[model.CredentialDocument] = None):  # type: (...) -> concurrent.futures.Future[model.GetClientDeviceAuthTokenResponse]
        """
        Perform the GetClientDeviceAuthToken operation asynchronously.

        Args:
            credential: 
        """
        request = model.GetClientDeviceAuthTokenRequest(credential=credential)
        operation = self.client.new_get_client_device_auth_token()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def get_component_details(self, *,
        component_name: typing.Optional[str] = None) -> model.GetComponentDetailsResponse:
        """
        Perform the GetComponentDetails operation synchronously.

        Args:
            component_name: 
        """
        return self.get_component_details_async(component_name=component_name).result()

    def get_component_details_async(self, *,
        component_name: typing.Optional[str] = None):  # type: (...) -> concurrent.futures.Future[model.GetComponentDetailsResponse]
        """
        Perform the GetComponentDetails operation asynchronously.

        Args:
            component_name: 
        """
        request = model.GetComponentDetailsRequest(component_name=component_name)
        operation = self.client.new_get_component_details()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def get_configuration(self, *,
        component_name: typing.Optional[str] = None,
        key_path: typing.Optional[typing.List[str]] = None) -> model.GetConfigurationResponse:
        """
        Perform the GetConfiguration operation synchronously.

        Args:
            component_name: 
            key_path: 
        """
        return self.get_configuration_async(component_name=component_name, key_path=key_path).result()

    def get_configuration_async(self, *,
        component_name: typing.Optional[str] = None,
        key_path: typing.Optional[typing.List[str]] = None):  # type: (...) -> concurrent.futures.Future[model.GetConfigurationResponse]
        """
        Perform the GetConfiguration operation asynchronously.

        Args:
            component_name: 
            key_path: 
        """
        request = model.GetConfigurationRequest(component_name=component_name, key_path=key_path)
        operation = self.client.new_get_configuration()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def get_local_deployment_status(self, *,
        deployment_id: typing.Optional[str] = None) -> model.GetLocalDeploymentStatusResponse:
        """
        Perform the GetLocalDeploymentStatus operation synchronously.

        Args:
            deployment_id: 
        """
        return self.get_local_deployment_status_async(deployment_id=deployment_id).result()

    def get_local_deployment_status_async(self, *,
        deployment_id: typing.Optional[str] = None):  # type: (...) -> concurrent.futures.Future[model.GetLocalDeploymentStatusResponse]
        """
        Perform the GetLocalDeploymentStatus operation asynchronously.

        Args:
            deployment_id: 
        """
        request = model.GetLocalDeploymentStatusRequest(deployment_id=deployment_id)
        operation = self.client.new_get_local_deployment_status()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def get_secret_value(self, *,
        secret_id: typing.Optional[str] = None,
        version_id: typing.Optional[str] = None,
        version_stage: typing.Optional[str] = None) -> model.GetSecretValueResponse:
        """
        Perform the GetSecretValue operation synchronously.

        Args:
            secret_id: 
            version_id: 
            version_stage: 
        """
        return self.get_secret_value_async(secret_id=secret_id, version_id=version_id, version_stage=version_stage).result()

    def get_secret_value_async(self, *,
        secret_id: typing.Optional[str] = None,
        version_id: typing.Optional[str] = None,
        version_stage: typing.Optional[str] = None):  # type: (...) -> concurrent.futures.Future[model.GetSecretValueResponse]
        """
        Perform the GetSecretValue operation asynchronously.

        Args:
            secret_id: 
            version_id: 
            version_stage: 
        """
        request = model.GetSecretValueRequest(secret_id=secret_id, version_id=version_id, version_stage=version_stage)
        operation = self.client.new_get_secret_value()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def get_thing_shadow(self, *,
        thing_name: typing.Optional[str] = None,
        shadow_name: typing.Optional[str] = None) -> model.GetThingShadowResponse:
        """
        Perform the GetThingShadow operation synchronously.

        Args:
            thing_name: 
            shadow_name: 
        """
        return self.get_thing_shadow_async(thing_name=thing_name, shadow_name=shadow_name).result()

    def get_thing_shadow_async(self, *,
        thing_name: typing.Optional[str] = None,
        shadow_name: typing.Optional[str] = None):  # type: (...) -> concurrent.futures.Future[model.GetThingShadowResponse]
        """
        Perform the GetThingShadow operation asynchronously.

        Args:
            thing_name: 
            shadow_name: 
        """
        request = model.GetThingShadowRequest(thing_name=thing_name, shadow_name=shadow_name)
        operation = self.client.new_get_thing_shadow()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def list_components(self) -> model.ListComponentsResponse:
        """
        Perform the ListComponents operation synchronously.

        """
        return self.list_components_async().result()

    def list_components_async(self):  # type: (...) -> concurrent.futures.Future[model.ListComponentsResponse]
        """
        Perform the ListComponents operation asynchronously.

        """
        request = model.ListComponentsRequest()
        operation = self.client.new_list_components()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def list_local_deployments(self) -> model.ListLocalDeploymentsResponse:
        """
        Perform the ListLocalDeployments operation synchronously.

        """
        return self.list_local_deployments_async().result()

    def list_local_deployments_async(self):  # type: (...) -> concurrent.futures.Future[model.ListLocalDeploymentsResponse]
        """
        Perform the ListLocalDeployments operation asynchronously.

        """
        request = model.ListLocalDeploymentsRequest()
        operation = self.client.new_list_local_deployments()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def list_named_shadows_for_thing(self, *,
        thing_name: typing.Optional[str] = None,
        next_token: typing.Optional[str] = None,
        page_size: typing.Optional[int] = None) -> model.ListNamedShadowsForThingResponse:
        """
        Perform the ListNamedShadowsForThing operation synchronously.

        Args:
            thing_name: 
            next_token: 
            page_size: 
        """
        return self.list_named_shadows_for_thing_async(thing_name=thing_name, next_token=next_token, page_size=page_size).result()

    def list_named_shadows_for_thing_async(self, *,
        thing_name: typing.Optional[str] = None,
        next_token: typing.Optional[str] = None,
        page_size: typing.Optional[int] = None):  # type: (...) -> concurrent.futures.Future[model.ListNamedShadowsForThingResponse]
        """
        Perform the ListNamedShadowsForThing operation asynchronously.

        Args:
            thing_name: 
            next_token: 
            page_size: 
        """
        request = model.ListNamedShadowsForThingRequest(thing_name=thing_name, next_token=next_token, page_size=page_size)
        operation = self.client.new_list_named_shadows_for_thing()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def pause_component(self, *,
        component_name: typing.Optional[str] = None) -> model.PauseComponentResponse:
        """
        Perform the PauseComponent operation synchronously.

        Args:
            component_name: 
        """
        return self.pause_component_async(component_name=component_name).result()

    def pause_component_async(self, *,
        component_name: typing.Optional[str] = None):  # type: (...) -> concurrent.futures.Future[model.PauseComponentResponse]
        """
        Perform the PauseComponent operation asynchronously.

        Args:
            component_name: 
        """
        request = model.PauseComponentRequest(component_name=component_name)
        operation = self.client.new_pause_component()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def publish_to_iot_core(self, *,
        topic_name: typing.Optional[str] = None,
        qos: typing.Optional[str] = None,
        payload: typing.Optional[typing.Union[bytes, str]] = None,
        retain: typing.Optional[bool] = None,
        user_properties: typing.Optional[typing.List[model.UserProperty]] = None,
        message_expiry_interval_seconds: typing.Optional[int] = None,
        correlation_data: typing.Optional[typing.Union[bytes, str]] = None,
        response_topic: typing.Optional[str] = None,
        payload_format: typing.Optional[str] = None,
        content_type: typing.Optional[str] = None) -> model.PublishToIoTCoreResponse:
        """
        Perform the PublishToIoTCore operation synchronously.

        Args:
            topic_name: 
            qos: QOS enum value
            payload: 
            retain: 
            user_properties: 
            message_expiry_interval_seconds: 
            correlation_data: 
            response_topic: 
            payload_format: PayloadFormat enum value
            content_type: 
        """
        return self.publish_to_iot_core_async(topic_name=topic_name, qos=qos, payload=payload, retain=retain, user_properties=user_properties, message_expiry_interval_seconds=message_expiry_interval_seconds, correlation_data=correlation_data, response_topic=response_topic, payload_format=payload_format, content_type=content_type).result()

    def publish_to_iot_core_async(self, *,
        topic_name: typing.Optional[str] = None,
        qos: typing.Optional[str] = None,
        payload: typing.Optional[typing.Union[bytes, str]] = None,
        retain: typing.Optional[bool] = None,
        user_properties: typing.Optional[typing.List[model.UserProperty]] = None,
        message_expiry_interval_seconds: typing.Optional[int] = None,
        correlation_data: typing.Optional[typing.Union[bytes, str]] = None,
        response_topic: typing.Optional[str] = None,
        payload_format: typing.Optional[str] = None,
        content_type: typing.Optional[str] = None):  # type: (...) -> concurrent.futures.Future[model.PublishToIoTCoreResponse]
        """
        Perform the PublishToIoTCore operation asynchronously.

        Args:
            topic_name: 
            qos: QOS enum value
            payload: 
            retain: 
            user_properties: 
            message_expiry_interval_seconds: 
            correlation_data: 
            response_topic: 
            payload_format: PayloadFormat enum value
            content_type: 
        """
        request = model.PublishToIoTCoreRequest(topic_name=topic_name, qos=qos, payload=payload, retain=retain, user_properties=user_properties, message_expiry_interval_seconds=message_expiry_interval_seconds, correlation_data=correlation_data, response_topic=response_topic, payload_format=payload_format, content_type=content_type)
        operation = self.client.new_publish_to_iot_core()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def publish_to_topic(self, *,
        topic: typing.Optional[str] = None,
        publish_message: typing.Optional[model.PublishMessage] = None) -> model.PublishToTopicResponse:
        """
        Perform the PublishToTopic operation synchronously.

        Args:
            topic: 
            publish_message: 
        """
        return self.publish_to_topic_async(topic=topic, publish_message=publish_message).result()

    def publish_to_topic_async(self, *,
        topic: typing.Optional[str] = None,
        publish_message: typing.Optional[model.PublishMessage] = None):  # type: (...) -> concurrent.futures.Future[model.PublishToTopicResponse]
        """
        Perform the PublishToTopic operation asynchronously.

        Args:
            topic: 
            publish_message: 
        """
        request = model.PublishToTopicRequest(topic=topic, publish_message=publish_message)
        operation = self.client.new_publish_to_topic()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def put_component_metric(self, *,
        metrics: typing.Optional[typing.List[model.Metric]] = None) -> model.PutComponentMetricResponse:
        """
        Perform the PutComponentMetric operation synchronously.

        Args:
            metrics: 
        """
        return self.put_component_metric_async(metrics=metrics).result()

    def put_component_metric_async(self, *,
        metrics: typing.Optional[typing.List[model.Metric]] = None):  # type: (...) -> concurrent.futures.Future[model.PutComponentMetricResponse]
        """
        Perform the PutComponentMetric operation asynchronously.

        Args:
            metrics: 
        """
        request = model.PutComponentMetricRequest(metrics=metrics)
        operation = self.client.new_put_component_metric()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def restart_component(self, *,
        component_name: typing.Optional[str] = None) -> model.RestartComponentResponse:
        """
        Perform the RestartComponent operation synchronously.

        Args:
            component_name: 
        """
        return self.restart_component_async(component_name=component_name).result()

    def restart_component_async(self, *,
        component_name: typing.Optional[str] = None):  # type: (...) -> concurrent.futures.Future[model.RestartComponentResponse]
        """
        Perform the RestartComponent operation asynchronously.

        Args:
            component_name: 
        """
        request = model.RestartComponentRequest(component_name=component_name)
        operation = self.client.new_restart_component()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def resume_component(self, *,
        component_name: typing.Optional[str] = None) -> model.ResumeComponentResponse:
        """
        Perform the ResumeComponent operation synchronously.

        Args:
            component_name: 
        """
        return self.resume_component_async(component_name=component_name).result()

    def resume_component_async(self, *,
        component_name: typing.Optional[str] = None):  # type: (...) -> concurrent.futures.Future[model.ResumeComponentResponse]
        """
        Perform the ResumeComponent operation asynchronously.

        Args:
            component_name: 
        """
        request = model.ResumeComponentRequest(component_name=component_name)
        operation = self.client.new_resume_component()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def send_configuration_validity_report(self, *,
        configuration_validity_report: typing.Optional[model.ConfigurationValidityReport] = None) -> model.SendConfigurationValidityReportResponse:
        """
        Perform the SendConfigurationValidityReport operation synchronously.

        Args:
            configuration_validity_report: 
        """
        return self.send_configuration_validity_report_async(configuration_validity_report=configuration_validity_report).result()

    def send_configuration_validity_report_async(self, *,
        configuration_validity_report: typing.Optional[model.ConfigurationValidityReport] = None):  # type: (...) -> concurrent.futures.Future[model.SendConfigurationValidityReportResponse]
        """
        Perform the SendConfigurationValidityReport operation asynchronously.

        Args:
            configuration_validity_report: 
        """
        request = model.SendConfigurationValidityReportRequest(configuration_validity_report=configuration_validity_report)
        operation = self.client.new_send_configuration_validity_report()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def stop_component(self, *,
        component_name: typing.Optional[str] = None) -> model.StopComponentResponse:
        """
        Perform the StopComponent operation synchronously.

        Args:
            component_name: 
        """
        return self.stop_component_async(component_name=component_name).result()

    def stop_component_async(self, *,
        component_name: typing.Optional[str] = None):  # type: (...) -> concurrent.futures.Future[model.StopComponentResponse]
        """
        Perform the StopComponent operation asynchronously.

        Args:
            component_name: 
        """
        request = model.StopComponentRequest(component_name=component_name)
        operation = self.client.new_stop_component()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def subscribe_to_certificate_updates(self, *,
        certificate_options: typing.Optional[model.CertificateOptions] = None,
        stream_handler: typing.Optional[client.SubscribeToCertificateUpdatesStreamHandler] = None,
        on_stream_event: typing.Optional[typing.Callable[[model.CertificateUpdateEvent], None]] = None,
        on_stream_error: typing.Optional[typing.Callable[[Exception], bool]] = None,
        on_stream_closed: typing.Optional[typing.Callable[[], None]] = None
) -> typing.Tuple[model.SubscribeToCertificateUpdatesResponse, client.SubscribeToCertificateUpdatesOperation]:
        """
        Perform the SubscribeToCertificateUpdates operation synchronously.
        The initial response or error will be returned synchronously, further events will arrive via the streaming
        callbacks

        Args:
            certificate_options: 
            stream_handler: Methods on this object will be called as stream events happen on this operation. If an
                executor is provided, the on_stream_event and on_stream_closed methods will run in the executor.
            on_stream_event: Callback for stream events. Mutually exclusive with stream_handler. If an executor is
                provided, this method will run in the executor.
            on_stream_error: Callback for stream errors. Return true to close the stream, return false to keep the
                stream open. Mutually exclusive with stream_handler. Even if an executor is provided, this method
                will not run in the executor.
            on_stream_closed: Callback for when the stream closes. Mutually exclusive with stream_handler. If an
                executor is provided, this method will run in the executor.
        """
        (fut, op) = self.subscribe_to_certificate_updates_async(certificate_options=certificate_options, 
            stream_handler=stream_handler, on_stream_event=on_stream_event, on_stream_error=on_stream_error,
            on_stream_closed=on_stream_closed)
        return fut.result(), op

    def subscribe_to_certificate_updates_async(self, *,
        certificate_options: typing.Optional[model.CertificateOptions] = None,
        stream_handler: client.SubscribeToCertificateUpdatesStreamHandler = None,
        on_stream_event: typing.Optional[typing.Callable[[model.CertificateUpdateEvent], None]] = None,
        on_stream_error: typing.Optional[typing.Callable[[Exception], bool]] = None,
        on_stream_closed: typing.Optional[typing.Callable[[], None]] = None
        ):  # type: (...) -> typing.Tuple[concurrent.futures.Future[model.SubscribeToCertificateUpdatesResponse], client.SubscribeToCertificateUpdatesOperation]
        """
        Perform the SubscribeToCertificateUpdates operation asynchronously.
        The initial response or error will be returned as the result of the asynchronous future, further events will
        arrive via the streaming callbacks

        Args:
            certificate_options: 
            stream_handler: Methods on this object will be called as stream events happen on this operation. If an
                executor is provided, the on_stream_event and on_stream_closed methods will run in the executor.
            on_stream_event: Callback for stream events. Mutually exclusive with stream_handler. If an executor is
                provided, this method will run in the executor.
            on_stream_error: Callback for stream errors. Return true to close the stream, return false to keep the
                stream open. Mutually exclusive with stream_handler. Even if an executor is provided, this method
                will not run in the executor.
            on_stream_closed: Callback for when the stream closes. Mutually exclusive with stream_handler. If an
                executor is provided, this method will run in the executor.
        """
        stream_handler = self.__handle_stream_handler("SubscribeToCertificateUpdates", stream_handler,
            on_stream_event, on_stream_error, on_stream_closed)
        request = model.SubscribeToCertificateUpdatesRequest(certificate_options=certificate_options)
        operation = self.client.new_subscribe_to_certificate_updates(stream_handler)
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response()), operation

    def subscribe_to_component_updates(self, *,
        stream_handler: typing.Optional[client.SubscribeToComponentUpdatesStreamHandler] = None,
        on_stream_event: typing.Optional[typing.Callable[[model.ComponentUpdatePolicyEvents], None]] = None,
        on_stream_error: typing.Optional[typing.Callable[[Exception], bool]] = None,
        on_stream_closed: typing.Optional[typing.Callable[[], None]] = None
) -> typing.Tuple[model.SubscribeToComponentUpdatesResponse, client.SubscribeToComponentUpdatesOperation]:
        """
        Perform the SubscribeToComponentUpdates operation synchronously.
        The initial response or error will be returned synchronously, further events will arrive via the streaming
        callbacks

        Args:
            stream_handler: Methods on this object will be called as stream events happen on this operation. If an
                executor is provided, the on_stream_event and on_stream_closed methods will run in the executor.
            on_stream_event: Callback for stream events. Mutually exclusive with stream_handler. If an executor is
                provided, this method will run in the executor.
            on_stream_error: Callback for stream errors. Return true to close the stream, return false to keep the
                stream open. Mutually exclusive with stream_handler. Even if an executor is provided, this method
                will not run in the executor.
            on_stream_closed: Callback for when the stream closes. Mutually exclusive with stream_handler. If an
                executor is provided, this method will run in the executor.
        """
        (fut, op) = self.subscribe_to_component_updates_async(
            stream_handler=stream_handler, on_stream_event=on_stream_event, on_stream_error=on_stream_error,
            on_stream_closed=on_stream_closed)
        return fut.result(), op

    def subscribe_to_component_updates_async(self, *,
        stream_handler: client.SubscribeToComponentUpdatesStreamHandler = None,
        on_stream_event: typing.Optional[typing.Callable[[model.ComponentUpdatePolicyEvents], None]] = None,
        on_stream_error: typing.Optional[typing.Callable[[Exception], bool]] = None,
        on_stream_closed: typing.Optional[typing.Callable[[], None]] = None
        ):  # type: (...) -> typing.Tuple[concurrent.futures.Future[model.SubscribeToComponentUpdatesResponse], client.SubscribeToComponentUpdatesOperation]
        """
        Perform the SubscribeToComponentUpdates operation asynchronously.
        The initial response or error will be returned as the result of the asynchronous future, further events will
        arrive via the streaming callbacks

        Args:
            stream_handler: Methods on this object will be called as stream events happen on this operation. If an
                executor is provided, the on_stream_event and on_stream_closed methods will run in the executor.
            on_stream_event: Callback for stream events. Mutually exclusive with stream_handler. If an executor is
                provided, this method will run in the executor.
            on_stream_error: Callback for stream errors. Return true to close the stream, return false to keep the
                stream open. Mutually exclusive with stream_handler. Even if an executor is provided, this method
                will not run in the executor.
            on_stream_closed: Callback for when the stream closes. Mutually exclusive with stream_handler. If an
                executor is provided, this method will run in the executor.
        """
        stream_handler = self.__handle_stream_handler("SubscribeToComponentUpdates", stream_handler,
            on_stream_event, on_stream_error, on_stream_closed)
        request = model.SubscribeToComponentUpdatesRequest()
        operation = self.client.new_subscribe_to_component_updates(stream_handler)
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response()), operation

    def subscribe_to_configuration_update(self, *,
        component_name: typing.Optional[str] = None,
        key_path: typing.Optional[typing.List[str]] = None,
        stream_handler: typing.Optional[client.SubscribeToConfigurationUpdateStreamHandler] = None,
        on_stream_event: typing.Optional[typing.Callable[[model.ConfigurationUpdateEvents], None]] = None,
        on_stream_error: typing.Optional[typing.Callable[[Exception], bool]] = None,
        on_stream_closed: typing.Optional[typing.Callable[[], None]] = None
) -> typing.Tuple[model.SubscribeToConfigurationUpdateResponse, client.SubscribeToConfigurationUpdateOperation]:
        """
        Perform the SubscribeToConfigurationUpdate operation synchronously.
        The initial response or error will be returned synchronously, further events will arrive via the streaming
        callbacks

        Args:
            component_name: 
            key_path: 
            stream_handler: Methods on this object will be called as stream events happen on this operation. If an
                executor is provided, the on_stream_event and on_stream_closed methods will run in the executor.
            on_stream_event: Callback for stream events. Mutually exclusive with stream_handler. If an executor is
                provided, this method will run in the executor.
            on_stream_error: Callback for stream errors. Return true to close the stream, return false to keep the
                stream open. Mutually exclusive with stream_handler. Even if an executor is provided, this method
                will not run in the executor.
            on_stream_closed: Callback for when the stream closes. Mutually exclusive with stream_handler. If an
                executor is provided, this method will run in the executor.
        """
        (fut, op) = self.subscribe_to_configuration_update_async(component_name=component_name, key_path=key_path, 
            stream_handler=stream_handler, on_stream_event=on_stream_event, on_stream_error=on_stream_error,
            on_stream_closed=on_stream_closed)
        return fut.result(), op

    def subscribe_to_configuration_update_async(self, *,
        component_name: typing.Optional[str] = None,
        key_path: typing.Optional[typing.List[str]] = None,
        stream_handler: client.SubscribeToConfigurationUpdateStreamHandler = None,
        on_stream_event: typing.Optional[typing.Callable[[model.ConfigurationUpdateEvents], None]] = None,
        on_stream_error: typing.Optional[typing.Callable[[Exception], bool]] = None,
        on_stream_closed: typing.Optional[typing.Callable[[], None]] = None
        ):  # type: (...) -> typing.Tuple[concurrent.futures.Future[model.SubscribeToConfigurationUpdateResponse], client.SubscribeToConfigurationUpdateOperation]
        """
        Perform the SubscribeToConfigurationUpdate operation asynchronously.
        The initial response or error will be returned as the result of the asynchronous future, further events will
        arrive via the streaming callbacks

        Args:
            component_name: 
            key_path: 
            stream_handler: Methods on this object will be called as stream events happen on this operation. If an
                executor is provided, the on_stream_event and on_stream_closed methods will run in the executor.
            on_stream_event: Callback for stream events. Mutually exclusive with stream_handler. If an executor is
                provided, this method will run in the executor.
            on_stream_error: Callback for stream errors. Return true to close the stream, return false to keep the
                stream open. Mutually exclusive with stream_handler. Even if an executor is provided, this method
                will not run in the executor.
            on_stream_closed: Callback for when the stream closes. Mutually exclusive with stream_handler. If an
                executor is provided, this method will run in the executor.
        """
        stream_handler = self.__handle_stream_handler("SubscribeToConfigurationUpdate", stream_handler,
            on_stream_event, on_stream_error, on_stream_closed)
        request = model.SubscribeToConfigurationUpdateRequest(component_name=component_name, key_path=key_path)
        operation = self.client.new_subscribe_to_configuration_update(stream_handler)
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response()), operation

    def subscribe_to_iot_core(self, *,
        topic_name: typing.Optional[str] = None,
        qos: typing.Optional[str] = None,
        stream_handler: typing.Optional[client.SubscribeToIoTCoreStreamHandler] = None,
        on_stream_event: typing.Optional[typing.Callable[[model.IoTCoreMessage], None]] = None,
        on_stream_error: typing.Optional[typing.Callable[[Exception], bool]] = None,
        on_stream_closed: typing.Optional[typing.Callable[[], None]] = None
) -> typing.Tuple[model.SubscribeToIoTCoreResponse, client.SubscribeToIoTCoreOperation]:
        """
        Perform the SubscribeToIoTCore operation synchronously.
        The initial response or error will be returned synchronously, further events will arrive via the streaming
        callbacks

        Args:
            topic_name: 
            qos: QOS enum value
            stream_handler: Methods on this object will be called as stream events happen on this operation. If an
                executor is provided, the on_stream_event and on_stream_closed methods will run in the executor.
            on_stream_event: Callback for stream events. Mutually exclusive with stream_handler. If an executor is
                provided, this method will run in the executor.
            on_stream_error: Callback for stream errors. Return true to close the stream, return false to keep the
                stream open. Mutually exclusive with stream_handler. Even if an executor is provided, this method
                will not run in the executor.
            on_stream_closed: Callback for when the stream closes. Mutually exclusive with stream_handler. If an
                executor is provided, this method will run in the executor.
        """
        (fut, op) = self.subscribe_to_iot_core_async(topic_name=topic_name, qos=qos, 
            stream_handler=stream_handler, on_stream_event=on_stream_event, on_stream_error=on_stream_error,
            on_stream_closed=on_stream_closed)
        return fut.result(), op

    def subscribe_to_iot_core_async(self, *,
        topic_name: typing.Optional[str] = None,
        qos: typing.Optional[str] = None,
        stream_handler: client.SubscribeToIoTCoreStreamHandler = None,
        on_stream_event: typing.Optional[typing.Callable[[model.IoTCoreMessage], None]] = None,
        on_stream_error: typing.Optional[typing.Callable[[Exception], bool]] = None,
        on_stream_closed: typing.Optional[typing.Callable[[], None]] = None
        ):  # type: (...) -> typing.Tuple[concurrent.futures.Future[model.SubscribeToIoTCoreResponse], client.SubscribeToIoTCoreOperation]
        """
        Perform the SubscribeToIoTCore operation asynchronously.
        The initial response or error will be returned as the result of the asynchronous future, further events will
        arrive via the streaming callbacks

        Args:
            topic_name: 
            qos: QOS enum value
            stream_handler: Methods on this object will be called as stream events happen on this operation. If an
                executor is provided, the on_stream_event and on_stream_closed methods will run in the executor.
            on_stream_event: Callback for stream events. Mutually exclusive with stream_handler. If an executor is
                provided, this method will run in the executor.
            on_stream_error: Callback for stream errors. Return true to close the stream, return false to keep the
                stream open. Mutually exclusive with stream_handler. Even if an executor is provided, this method
                will not run in the executor.
            on_stream_closed: Callback for when the stream closes. Mutually exclusive with stream_handler. If an
                executor is provided, this method will run in the executor.
        """
        stream_handler = self.__handle_stream_handler("SubscribeToIoTCore", stream_handler,
            on_stream_event, on_stream_error, on_stream_closed)
        request = model.SubscribeToIoTCoreRequest(topic_name=topic_name, qos=qos)
        operation = self.client.new_subscribe_to_iot_core(stream_handler)
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response()), operation

    def subscribe_to_topic(self, *,
        topic: typing.Optional[str] = None,
        receive_mode: typing.Optional[str] = None,
        stream_handler: typing.Optional[client.SubscribeToTopicStreamHandler] = None,
        on_stream_event: typing.Optional[typing.Callable[[model.SubscriptionResponseMessage], None]] = None,
        on_stream_error: typing.Optional[typing.Callable[[Exception], bool]] = None,
        on_stream_closed: typing.Optional[typing.Callable[[], None]] = None
) -> typing.Tuple[model.SubscribeToTopicResponse, client.SubscribeToTopicOperation]:
        """
        Perform the SubscribeToTopic operation synchronously.
        The initial response or error will be returned synchronously, further events will arrive via the streaming
        callbacks

        Args:
            topic: 
            receive_mode: ReceiveMode enum value
            stream_handler: Methods on this object will be called as stream events happen on this operation. If an
                executor is provided, the on_stream_event and on_stream_closed methods will run in the executor.
            on_stream_event: Callback for stream events. Mutually exclusive with stream_handler. If an executor is
                provided, this method will run in the executor.
            on_stream_error: Callback for stream errors. Return true to close the stream, return false to keep the
                stream open. Mutually exclusive with stream_handler. Even if an executor is provided, this method
                will not run in the executor.
            on_stream_closed: Callback for when the stream closes. Mutually exclusive with stream_handler. If an
                executor is provided, this method will run in the executor.
        """
        (fut, op) = self.subscribe_to_topic_async(topic=topic, receive_mode=receive_mode, 
            stream_handler=stream_handler, on_stream_event=on_stream_event, on_stream_error=on_stream_error,
            on_stream_closed=on_stream_closed)
        return fut.result(), op

    def subscribe_to_topic_async(self, *,
        topic: typing.Optional[str] = None,
        receive_mode: typing.Optional[str] = None,
        stream_handler: client.SubscribeToTopicStreamHandler = None,
        on_stream_event: typing.Optional[typing.Callable[[model.SubscriptionResponseMessage], None]] = None,
        on_stream_error: typing.Optional[typing.Callable[[Exception], bool]] = None,
        on_stream_closed: typing.Optional[typing.Callable[[], None]] = None
        ):  # type: (...) -> typing.Tuple[concurrent.futures.Future[model.SubscribeToTopicResponse], client.SubscribeToTopicOperation]
        """
        Perform the SubscribeToTopic operation asynchronously.
        The initial response or error will be returned as the result of the asynchronous future, further events will
        arrive via the streaming callbacks

        Args:
            topic: 
            receive_mode: ReceiveMode enum value
            stream_handler: Methods on this object will be called as stream events happen on this operation. If an
                executor is provided, the on_stream_event and on_stream_closed methods will run in the executor.
            on_stream_event: Callback for stream events. Mutually exclusive with stream_handler. If an executor is
                provided, this method will run in the executor.
            on_stream_error: Callback for stream errors. Return true to close the stream, return false to keep the
                stream open. Mutually exclusive with stream_handler. Even if an executor is provided, this method
                will not run in the executor.
            on_stream_closed: Callback for when the stream closes. Mutually exclusive with stream_handler. If an
                executor is provided, this method will run in the executor.
        """
        stream_handler = self.__handle_stream_handler("SubscribeToTopic", stream_handler,
            on_stream_event, on_stream_error, on_stream_closed)
        request = model.SubscribeToTopicRequest(topic=topic, receive_mode=receive_mode)
        operation = self.client.new_subscribe_to_topic(stream_handler)
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response()), operation

    def subscribe_to_validate_configuration_updates(self, *,
        stream_handler: typing.Optional[client.SubscribeToValidateConfigurationUpdatesStreamHandler] = None,
        on_stream_event: typing.Optional[typing.Callable[[model.ValidateConfigurationUpdateEvents], None]] = None,
        on_stream_error: typing.Optional[typing.Callable[[Exception], bool]] = None,
        on_stream_closed: typing.Optional[typing.Callable[[], None]] = None
) -> typing.Tuple[model.SubscribeToValidateConfigurationUpdatesResponse, client.SubscribeToValidateConfigurationUpdatesOperation]:
        """
        Perform the SubscribeToValidateConfigurationUpdates operation synchronously.
        The initial response or error will be returned synchronously, further events will arrive via the streaming
        callbacks

        Args:
            stream_handler: Methods on this object will be called as stream events happen on this operation. If an
                executor is provided, the on_stream_event and on_stream_closed methods will run in the executor.
            on_stream_event: Callback for stream events. Mutually exclusive with stream_handler. If an executor is
                provided, this method will run in the executor.
            on_stream_error: Callback for stream errors. Return true to close the stream, return false to keep the
                stream open. Mutually exclusive with stream_handler. Even if an executor is provided, this method
                will not run in the executor.
            on_stream_closed: Callback for when the stream closes. Mutually exclusive with stream_handler. If an
                executor is provided, this method will run in the executor.
        """
        (fut, op) = self.subscribe_to_validate_configuration_updates_async(
            stream_handler=stream_handler, on_stream_event=on_stream_event, on_stream_error=on_stream_error,
            on_stream_closed=on_stream_closed)
        return fut.result(), op

    def subscribe_to_validate_configuration_updates_async(self, *,
        stream_handler: client.SubscribeToValidateConfigurationUpdatesStreamHandler = None,
        on_stream_event: typing.Optional[typing.Callable[[model.ValidateConfigurationUpdateEvents], None]] = None,
        on_stream_error: typing.Optional[typing.Callable[[Exception], bool]] = None,
        on_stream_closed: typing.Optional[typing.Callable[[], None]] = None
        ):  # type: (...) -> typing.Tuple[concurrent.futures.Future[model.SubscribeToValidateConfigurationUpdatesResponse], client.SubscribeToValidateConfigurationUpdatesOperation]
        """
        Perform the SubscribeToValidateConfigurationUpdates operation asynchronously.
        The initial response or error will be returned as the result of the asynchronous future, further events will
        arrive via the streaming callbacks

        Args:
            stream_handler: Methods on this object will be called as stream events happen on this operation. If an
                executor is provided, the on_stream_event and on_stream_closed methods will run in the executor.
            on_stream_event: Callback for stream events. Mutually exclusive with stream_handler. If an executor is
                provided, this method will run in the executor.
            on_stream_error: Callback for stream errors. Return true to close the stream, return false to keep the
                stream open. Mutually exclusive with stream_handler. Even if an executor is provided, this method
                will not run in the executor.
            on_stream_closed: Callback for when the stream closes. Mutually exclusive with stream_handler. If an
                executor is provided, this method will run in the executor.
        """
        stream_handler = self.__handle_stream_handler("SubscribeToValidateConfigurationUpdates", stream_handler,
            on_stream_event, on_stream_error, on_stream_closed)
        request = model.SubscribeToValidateConfigurationUpdatesRequest()
        operation = self.client.new_subscribe_to_validate_configuration_updates(stream_handler)
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response()), operation

    def update_configuration(self, *,
        key_path: typing.Optional[typing.List[str]] = None,
        timestamp: typing.Optional[datetime.datetime] = None,
        value_to_merge: typing.Optional[typing.Dict[str, typing.Any]] = None) -> model.UpdateConfigurationResponse:
        """
        Perform the UpdateConfiguration operation synchronously.

        Args:
            key_path: 
            timestamp: 
            value_to_merge: 
        """
        return self.update_configuration_async(key_path=key_path, timestamp=timestamp, value_to_merge=value_to_merge).result()

    def update_configuration_async(self, *,
        key_path: typing.Optional[typing.List[str]] = None,
        timestamp: typing.Optional[datetime.datetime] = None,
        value_to_merge: typing.Optional[typing.Dict[str, typing.Any]] = None):  # type: (...) -> concurrent.futures.Future[model.UpdateConfigurationResponse]
        """
        Perform the UpdateConfiguration operation asynchronously.

        Args:
            key_path: 
            timestamp: 
            value_to_merge: 
        """
        request = model.UpdateConfigurationRequest(key_path=key_path, timestamp=timestamp, value_to_merge=value_to_merge)
        operation = self.client.new_update_configuration()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def update_state(self, *,
        state: typing.Optional[str] = None) -> model.UpdateStateResponse:
        """
        Perform the UpdateState operation synchronously.

        Args:
            state: ReportedLifecycleState enum value
        """
        return self.update_state_async(state=state).result()

    def update_state_async(self, *,
        state: typing.Optional[str] = None):  # type: (...) -> concurrent.futures.Future[model.UpdateStateResponse]
        """
        Perform the UpdateState operation asynchronously.

        Args:
            state: ReportedLifecycleState enum value
        """
        request = model.UpdateStateRequest(state=state)
        operation = self.client.new_update_state()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def update_thing_shadow(self, *,
        thing_name: typing.Optional[str] = None,
        shadow_name: typing.Optional[str] = None,
        payload: typing.Optional[typing.Union[bytes, str]] = None) -> model.UpdateThingShadowResponse:
        """
        Perform the UpdateThingShadow operation synchronously.

        Args:
            thing_name: 
            shadow_name: 
            payload: 
        """
        return self.update_thing_shadow_async(thing_name=thing_name, shadow_name=shadow_name, payload=payload).result()

    def update_thing_shadow_async(self, *,
        thing_name: typing.Optional[str] = None,
        shadow_name: typing.Optional[str] = None,
        payload: typing.Optional[typing.Union[bytes, str]] = None):  # type: (...) -> concurrent.futures.Future[model.UpdateThingShadowResponse]
        """
        Perform the UpdateThingShadow operation asynchronously.

        Args:
            thing_name: 
            shadow_name: 
            payload: 
        """
        request = model.UpdateThingShadowRequest(thing_name=thing_name, shadow_name=shadow_name, payload=payload)
        operation = self.client.new_update_thing_shadow()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def validate_authorization_token(self, *,
        token: typing.Optional[str] = None) -> model.ValidateAuthorizationTokenResponse:
        """
        Perform the ValidateAuthorizationToken operation synchronously.

        Args:
            token: 
        """
        return self.validate_authorization_token_async(token=token).result()

    def validate_authorization_token_async(self, *,
        token: typing.Optional[str] = None):  # type: (...) -> concurrent.futures.Future[model.ValidateAuthorizationTokenResponse]
        """
        Perform the ValidateAuthorizationToken operation asynchronously.

        Args:
            token: 
        """
        request = model.ValidateAuthorizationTokenRequest(token=token)
        operation = self.client.new_validate_authorization_token()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())

    def verify_client_device_identity(self, *,
        credential: typing.Optional[model.ClientDeviceCredential] = None) -> model.VerifyClientDeviceIdentityResponse:
        """
        Perform the VerifyClientDeviceIdentity operation synchronously.

        Args:
            credential: 
        """
        return self.verify_client_device_identity_async(credential=credential).result()

    def verify_client_device_identity_async(self, *,
        credential: typing.Optional[model.ClientDeviceCredential] = None):  # type: (...) -> concurrent.futures.Future[model.VerifyClientDeviceIdentityResponse]
        """
        Perform the VerifyClientDeviceIdentity operation asynchronously.

        Args:
            credential: 
        """
        request = model.VerifyClientDeviceIdentityRequest(credential=credential)
        operation = self.client.new_verify_client_device_identity()
        write_future = operation.activate(request)
        return self.__combine_futures(write_future, operation.get_response())



import concurrent.futures
import threading
from unittest import TestCase
from unittest.mock import patch
import io
import contextlib

from awsiot.greengrasscoreipc.client import SubscribeToTopicStreamHandler
from awsiot.greengrasscoreipc.model import CreateLocalDeploymentResponse, SubscribeToTopicResponse, \
    SubscriptionResponseMessage, BinaryMessage

TIMEOUT = 10.0 # seconds


class GGV2Test(TestCase):

    def _mock_operation(self, mock_op, response):
        activate_fut = concurrent.futures.Future()
        activate_fut.set_result(None)
        mock_op.activate.return_value = activate_fut
        response_fut = concurrent.futures.Future()
        response_fut.set_result(response)
        mock_op.get_response.return_value = response_fut
        return mock_op

    @patch('awsiot.greengrasscoreipc.client.GreengrassCoreIPCClient')
    @patch('awsiot.greengrasscoreipc.client.CreateLocalDeploymentOperation')
    @patch('awsiot.greengrasscoreipc.client.SubscribeToTopicOperation')
    def test_connect(self, mock_client, mock_deployment_op, mock_subscribe_op):
        from awsiot.greengrasscoreipc.clientv2 import GreengrassCoreIPCClientV2 as Client
        c = Client(client=mock_client)

        self._mock_operation(mock_deployment_op, CreateLocalDeploymentResponse(deployment_id="deployment"))
        mock_client.new_create_local_deployment.return_value = mock_deployment_op
        resp = c.create_local_deployment()
        self.assertEqual("deployment", resp.deployment_id)

        # Verify subscription works and callback is called on the executor thread
        self._mock_operation(mock_subscribe_op, SubscribeToTopicResponse(topic_name="abc"))
        mock_client.new_subscribe_to_topic.return_value = mock_subscribe_op

        subscription_fut = concurrent.futures.Future()
        thread_id_fut = concurrent.futures.Future()

        def on_stream_event(r):
            subscription_fut.set_result(r)
            thread_id_fut.set_result(threading.get_ident())
        resp, op = c.subscribe_to_topic(topic="abc", on_stream_event=on_stream_event)
        self.assertEqual("abc", resp.topic_name)

        sub_handler = mock_client.new_subscribe_to_topic.call_args[0][0]
        sub_handler.on_stream_event(SubscriptionResponseMessage(binary_message=BinaryMessage(message="xyz")))

        self.assertEqual("xyz".encode("utf-8"), subscription_fut.result(TIMEOUT).binary_message.message)
        self.assertNotEqual(threading.get_ident(), thread_id_fut.result(TIMEOUT))

        # Verify that when using the stream_handler option, the callback is run in the executor

        subscription_fut = concurrent.futures.Future()
        thread_id_fut = concurrent.futures.Future()

        class handler(SubscribeToTopicStreamHandler):
            def on_stream_event(self, event):
                on_stream_event(event)
        resp, op = c.subscribe_to_topic(topic="abc", stream_handler=handler())
        self.assertEqual("abc", resp.topic_name)

        sub_handler = mock_client.new_subscribe_to_topic.call_args[0][0]
        sub_handler.on_stream_event(SubscriptionResponseMessage(binary_message=BinaryMessage(message="xyz")))

        self.assertEqual("xyz".encode("utf-8"), subscription_fut.result(TIMEOUT).binary_message.message)
        self.assertNotEqual(threading.get_ident(), thread_id_fut.result(TIMEOUT))

        # Remove executor from client to verify that we are not running the callback in a different thread
        c = Client(client=mock_client, executor=None)

        subscription_fut = concurrent.futures.Future()
        thread_id_fut = concurrent.futures.Future()

        resp, op = c.subscribe_to_topic(topic="abc", on_stream_event=on_stream_event)
        self.assertEqual("abc", resp.topic_name)

        sub_handler = mock_client.new_subscribe_to_topic.call_args[0][0]
        sub_handler.on_stream_event(SubscriptionResponseMessage(binary_message=BinaryMessage(message="xyz")))

        self.assertEqual("xyz".encode("utf-8"), subscription_fut.result(TIMEOUT).binary_message.message)
        self.assertEqual(threading.get_ident(), thread_id_fut.result(TIMEOUT))

        # Verify we nicely print errors in user-provided handler methods
        def on_stream_event(r):
            raise ValueError("Broken!")

        c.subscribe_to_topic(topic="abc", on_stream_event=on_stream_event)
        sub_handler = mock_client.new_subscribe_to_topic.call_args[0][0]
        f = io.StringIO()
        with contextlib.redirect_stderr(f):
            self.assertRaises(ValueError, lambda: sub_handler.on_stream_event(
                SubscriptionResponseMessage(binary_message=BinaryMessage(message="xyz"))))
        self.assertIn("ValueError: Broken!", f.getvalue())



from pydoc import describe
from unittest import TestCase

from awsiot.iotshadow import ShadowState


class ShadowTest(TestCase):

    def test_shadow_state_payload_filled(self):
        test_state = ShadowState.from_payload({
            "reported": {"Color": "Red"},
            "desired": {"Color": "Blue"}
        })
        compareState = ShadowState(
            reported={"Color": "Red"},
            desired={"Color": "Blue"}
        )
        self.assertTrue(test_state.to_payload() == compareState.to_payload())

    def test_shadow_state_payload_partial_filled(self):
        test_state = ShadowState.from_payload({
            "reported": {"Color": "Red"}
        })
        self.assertEqual(test_state.to_payload(), {"reported": {"Color": "Red"}})

    def test_shadow_state_payload_can_send_null(self):
        test_state = ShadowState(
            reported={"Color": "Red"},
            desired=None,
            desired_is_nullable=True
        )
        self.assertEqual(test_state.to_payload(), {"reported": {"Color": "Red"}, "desired": None})

    def test_shadow_state_payload_with_none(self):
        test_state = ShadowState.from_payload({"reported": {"Color": "Red"}, "desired": None})
        self.assertTrue(test_state.desired_is_nullable)
        self.assertEqual(test_state.to_payload(), {"reported": {"Color": "Red"}, "desired": None})

    def test_shadow_state_payload_without_none(self):
        test_state = ShadowState.from_payload({"reported": {"deviceAgent": {"rebootRequired": True}}})
        self.assertEqual(test_state.to_payload(), {"reported": {"deviceAgent": {"rebootRequired": True}}})



import test.echotestrpc.model as model
import test.echotestrpc.client as client
from awsiot.eventstreamrpc import (
    AccessDeniedError,
    Connection,
    ConnectionClosedError,
    EventStreamError,
    LifecycleHandler,
    MessageAmendment,
    SerializeError,
    StreamClosedError,
    StreamResponseHandler)
from awscrt.io import (ClientBootstrap, DefaultHostResolver, EventLoopGroup,
                       init_logging, LogLevel)
from awscrt.eventstream import Header, HeaderType
from awscrt.eventstream.rpc import MessageType
from datetime import datetime, timezone
import logging
import os
from queue import Queue
from sys import stderr
from threading import Event
from time import sleep
from typing import Optional, Sequence
from unittest import skipUnless, TestCase

import awsiot.greengrasscoreipc.client

TIMEOUT = 10

#logging.basicConfig(level=logging.DEBUG, format='[%(levelname)s] [%(name)s] - %(message)s')
#init_logging(LogLevel.Trace, 'stderr')

EVENTSTREAM_ECHO_TEST = os.getenv('EVENTSTREAM_ECHO_TEST')


class ConnectionLifecycleHandler(LifecycleHandler):
    def __init__(self, on_freakout):
        self.connect_event = Event()
        self.disconnect_event = Event()
        self.disconnect_reason = None
        self.errors = Queue()
        self.error_callback_return_val = True
        self.pings = Queue()
        # if something happens out of order, call this
        self._freakout = on_freakout

    def on_connect(self):
        if self.disconnect_event.isSet():
            self._freakout("on_disconnect fired before on_connect()")
        if self.connect_event.isSet():
            self._freakout("on_connect fired multiple times")
        else:
            self.connect_event.set()

    def on_disconnect(self, reason: Optional[Exception]):
        if not self.connect_event.isSet():
            self._freakout("on_disconnect fired before on_connect")
        if self.disconnect_event.isSet():
            self._freakout("on_disconnect fired multiple times")
        else:
            if reason is not None and not isinstance(reason, Exception):
                self._freakout("on_disconnect reason is not an exception")
            self.disconnect_reason = reason
            self.disconnect_event.set()

    def on_error(self, error: Exception) -> bool:
        self.errors.put(error)
        return self.error_callback_return_val

    def on_ping(self, headers: Sequence[Header], payload: bytes):
        self.pings.put({'headers': headers, 'payload': payload})


class StreamHandler(StreamResponseHandler):
    def __init__(self, on_freakout):
        super().__init__()
        self.events = Queue()
        self.errors = Queue()
        self.closed = Event()
        self.error_callback_return_val = True
        # set this before activating operation
        self.operation = None
        # if something happens out of order, call this
        self._freakout = on_freakout

    def on_stream_event(self, event):
        if not self.operation.get_response().done():
            self._freakout("received event before initial response")
        if self.closed.is_set():
            self._freakout("received event after close")
        self.events.put(event)

    def on_stream_error(self, error: Exception) -> bool:
        if not self.operation.get_response().done():
            self._freakout("received event before initial response")
        if self.closed.is_set():
            self._freakout("received event after close")
        if not isinstance(error, Exception):
            self._freakout("on_stream_error delivered non-error")
        self.errors.put(error)
        return self.error_callback_return_val

    def on_stream_closed(self):
        if self.closed.is_set():
            self._freakout("received closed event twice")
        self.closed.set()


def connect_amender():
    headers = [Header.from_string('client-name', 'accepted.testy_mc_testerson')]
    return MessageAmendment(headers=headers)


@skipUnless(EVENTSTREAM_ECHO_TEST, "Skipping until we have permanent echo server")
class RpcTest(TestCase):
    def _on_handler_freakout(self, msg):
        print(msg, file=stderr)
        if not hasattr(self, 'freakout_msg'):
            self.freakout_msg = msg

    def _assertNoHandlerFreakout(self):
        self.assertIsNone(getattr(self, 'freakout_msg', None))

    def setUp(self):
        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)
        self.connection = Connection(
            host_name='127.0.0.1',
            port=8033,
            bootstrap=bootstrap,
            connect_message_amender=connect_amender)

    def _connect(self, lifecycle_handler=None):
        if lifecycle_handler:
            self.lifecycle_handler = lifecycle_handler
        else:
            self.lifecycle_handler = ConnectionLifecycleHandler(self._on_handler_freakout)
        connect_future = self.connection.connect(self.lifecycle_handler)
        connect_future.result(TIMEOUT)
        self.echo_client = client.EchoTestRPCClient(self.connection)

    def _bad_connect(self, bad_host=False, amender=None):
        if bad_host:
            self.connection.host_name = 'badhostname'
        if amender:
            self.connection._connect_message_amender = amender
        self.lifecycle_handler = ConnectionLifecycleHandler(self._on_handler_freakout)
        connect_future = self.connection.connect(self.lifecycle_handler)
        connect_exception = connect_future.exception(TIMEOUT)

        # connect attempt should fail
        self.assertIsNotNone(connect_exception)

        # no lifecycle events should have fired
        self.assertFalse(self.lifecycle_handler.connect_event.isSet())
        self.assertFalse(self.lifecycle_handler.disconnect_event.isSet())
        self.assertTrue(self.lifecycle_handler.errors.empty())

        self._assertNoHandlerFreakout()

        return connect_exception

    def _close_connection(self):
        # helper to do normal close of healthy connection
        close_future = self.connection.close()
        close_exception = close_future.exception(TIMEOUT)
        self.assertIsNone(close_exception)
        self.assertTrue(self.lifecycle_handler.disconnect_event.wait(TIMEOUT))
        self.assertIsNone(self.lifecycle_handler.disconnect_reason)
        self.assertTrue(self.lifecycle_handler.errors.empty())
        self._assertNoHandlerFreakout()

    def test_connect_failed_socket(self):
        # test failure from the CONNECTING_TO_SOCKET phase
        self._bad_connect(bad_host=True)

    def test_connect_failed_connack(self):
        # test failure from the WAITING_FOR_CONNECT_ACK phase
        def _amender():
            headers = [Header.from_string('client-name', 'rejected.testy_mc_failureson')]
            return MessageAmendment(headers=headers)
        exception = self._bad_connect(amender=_amender)
        self.assertIsInstance(exception, AccessDeniedError)

    def test_connect_failed_amender_exception(self):
        # test failure due to connect_amender exception
        error = RuntimeError('Purposefully raising error in amender callback')

        def _amender():
            raise error
        exception = self._bad_connect(amender=_amender)
        self.assertIs(exception, error)

    def test_connect_failed_amender_bad_return(self):
        # test failure due to amender returning bad data
        def _amender():
            return 'a string is not a MessageAmendment'
        self._bad_connect(amender=_amender)

    def test_echo_message(self):
        self._connect()

        operation = self.echo_client.new_echo_message()

        request = model.EchoMessageRequest(message=model.MessageData())
        request.message.string_message = '👁👄👁'
        request.message.boolean_message = True
        # hand-picked timestamp shouldn't lose precision when serialized to
        # json as floating point. also setting timezone to UTC because
        # response timezone will always be UTC
        request.message.time_message = datetime.fromtimestamp(946684800, tz=timezone.utc)
        request.message.document_message = {
            'inty': 4,
            'floaty': 4.0,
            'listy': [1, "two", 3.0, False, True, {}, None],
            'dicty': {'a': [1, 2, 3, {'e': 9}], 'b': None},
            'stringy': "qwer",
            'booly_t': True,
            'booly_f': False,
            'nully': None,
        }
        request.message.enum_message = model.FruitEnum.BANANA
        request.message.blob_message = bytes(list(range(256)))
        request.message.string_list_message = ['one', 'two', 'three']
        request.message.key_value_pair_list = [
            model.Pair(key='key1', value='value1'),
            model.Pair(key='key1', value='value1')]

        request_flush = operation.activate(request)
        self.assertIsNone(request_flush.result(TIMEOUT))

        response = operation.get_response().result(TIMEOUT)

        self.assertIsInstance(response, model.EchoMessageResponse)
        self.assertIsInstance(response.message, model.MessageData)
        # explicit tests for each member so it's clear exactly what went wrong
        self.assertEqual(request.message.string_message, response.message.string_message)
        self.assertEqual(request.message.boolean_message, response.message.boolean_message)
        self.assertEqual(request.message.document_message, response.message.document_message)
        self.assertEqual(request.message.enum_message, response.message.enum_message)
        self.assertEqual(request.message.blob_message, response.message.blob_message)
        self.assertEqual(request.message.string_list_message, response.message.string_list_message)
        self.assertEqual(request.message.key_value_pair_list, response.message.key_value_pair_list)
        self.assertIsInstance(response.message.time_message, datetime)
        self.assertAlmostEqual(request.message.time_message.timestamp(), response.message.time_message.timestamp())

        # ok now compare a whole class
        # if this fails, it's likely due to the datetime losing precision
        # and timezone info due to datetime->timestamp->datetime conversion
        self.assertEqual(request.message, response.message)

        self._close_connection()

    def test_bad_activate(self):
        self._connect()

        operation = self.echo_client.new_echo_message()

        bad_request = model.EchoMessageRequest()
        bad_request.message = ".message is not supposed to be a string"

        with self.assertRaises(SerializeError):
            operation.activate(bad_request)

        self._close_connection()

    def test_echo_stream_messages(self):
        self._connect()

        stream_handler = StreamHandler(self._on_handler_freakout)
        operation = self.echo_client.new_echo_stream_messages(stream_handler)
        stream_handler.operation = operation

        # send initial request
        flush = operation.activate(model.EchoStreamingRequest())
        flush.result(TIMEOUT)

        # send streaming request
        request_event = model.EchoStreamingMessage(key_value_pair=model.Pair(key='Kiki', value='Valerie'))
        flush = operation.send_stream_event(request_event)
        flush.result(TIMEOUT)

        # recv streaming response
        response_event = stream_handler.events.get(timeout=TIMEOUT)
        self.assertEqual(request_event, response_event)

        self._close_connection()
        self.assertTrue(stream_handler.closed.is_set())
        self.assertTrue(stream_handler.errors.empty())

    def test_cause_service_error(self):
        # test the CauseServiceError operation,
        # which always responds with a ServiceError
        self._connect()

        operation = self.echo_client.new_cause_service_error()

        # send initial request
        operation.activate(model.CauseServiceErrorRequest())

        # get response
        response_exception = operation.get_response().exception(TIMEOUT)
        self.assertIsInstance(response_exception, model.ServiceError)

        self._close_connection()

    def test_cause_stream_service_to_error(self):
        # test CauseStreamServiceToError operation,
        # Responds to initial request normally then throws a ServiceError on stream response
        self._connect()

        # set up operation
        stream_handler = StreamHandler(self._on_handler_freakout)
        stream_handler.error_callback_return_val = False
        op = self.echo_client.new_cause_stream_service_to_error(stream_handler)
        stream_handler.operation = op

        # send initial request, normal response should come back
        request = model.EchoStreamingRequest()
        op.activate(request)
        op.get_response().result()

        # send subsequent streaming message, streaming error should come back
        msg_to_send = model.EchoStreamingMessage(stream_message=model.MessageData())
        op.send_stream_event(msg_to_send)

        stream_error = stream_handler.errors.get(timeout=TIMEOUT)
        self.assertIsInstance(stream_error, model.ServiceError)

        self._close_connection()
        self.assertTrue(stream_handler.closed.is_set())
        self.assertTrue(stream_handler.errors.empty())

    def test_connection_error(self):
        # test that everything acts as expected if server sends
        # connection-level error
        self._connect()

        # reach deep into private inner workings of the connection to manually
        # send a bad message to the server.
        self.connection._synced.current_connection.send_protocol_message(
            headers=[Header.from_int32(':stream-id', -999)],
            message_type=MessageType.APPLICATION_MESSAGE,
        )

        # should receive PROTOCOL_ERROR in response to bad message
        error = self.lifecycle_handler.errors.get(timeout=TIMEOUT)
        self.assertIsInstance(error, EventStreamError)

        # server kills connection after PROTOCOL_ERROR
        self.assertTrue(self.lifecycle_handler.disconnect_event.wait(TIMEOUT))
        self.assertIsInstance(self.lifecycle_handler.disconnect_reason, EventStreamError)

    def test_close_with_reason(self):
        # test that, if an error is passed to connection.close(err),
        # it carries through
        self._connect()

        my_error = RuntimeError('my close reason')
        close_future = self.connection.close(my_error)
        close_reason = close_future.exception(TIMEOUT)

        self.assertIs(my_error, close_reason)
        self.assertTrue(self.lifecycle_handler.disconnect_event.wait(TIMEOUT))
        self.assertIs(my_error, self.lifecycle_handler.disconnect_reason)

    def test_reconnect(self):
        # test that a Connection can connect and disconnect multiple times
        self._connect()
        self._close_connection()

        self._connect()
        self._close_connection()

    def test_close_during_setup(self):
        # Test that it's safe to call close() while the connection is still setting up.

        # There are multiple stages to the async connect() and we'd like
        # to stress close() being called in each of these phases.
        # Hacky strategy to achieve this to, in a loop:
        # - call async connect()
        # - after some delay, call async close()
        # - with each loop, the delay gets slightly longer
        # - break out of loop loop once the delay is long enough that the
        #   connect() is completing before we ever call close()
        delay_increment_sec = 0.005
        stop_after_n_successful_connections = 2

        delay_sec = 0.0
        successful_connections = 0
        while successful_connections < stop_after_n_successful_connections:
            # not using helper _connect() call because it blocks until async connect() completes
            self.lifecycle_handler = ConnectionLifecycleHandler(self._on_handler_freakout)
            connect_future = self.connection.connect(self.lifecycle_handler)

            if delay_sec > 0.0:
                sleep(delay_sec)
            close_future = self.connection.close()

            # wait for connect and close to complete
            connect_exception = connect_future.exception(TIMEOUT)
            close_exception = close_future.exception(TIMEOUT)

            # close should have been clean
            self.assertIsNone(close_exception)

            # connect might have succeeded, or might have failed,
            # depending on the timing of this thread's close() call
            if connect_exception:
                self.assertIsInstance(connect_exception, ConnectionClosedError)
                # lifecycle handlers should NOT fire if connect setup failed
                # wait a tiny bit to be 100% sure these never fire
                self.assertFalse(self.lifecycle_handler.connect_event.wait(0.1))
                self.assertFalse(self.lifecycle_handler.disconnect_event.wait(0.1))
            else:
                self.assertTrue(self.lifecycle_handler.connect_event.wait(TIMEOUT))
                self.assertTrue(self.lifecycle_handler.disconnect_event.wait(TIMEOUT))
                successful_connections += 1

            delay_sec += delay_increment_sec
            self._assertNoHandlerFreakout()

    def test_operation_response_completes_if_connection_closed(self):
        # test that response future completes if connection is closed
        # before actual response is received.

        # this test is timing dependent, the response could theoretically
        # come on another thread before this thread can close the connection,
        # so run test in a loop till we get the timing we want
        closed_before_response = False

        # give up after a reasonably high number of tries
        # (note: first try always passes on my 2019 macbook pro, with localhost server)
        tries = 0
        max_tries = 100

        while not closed_before_response:
            self.assertLess(tries, max_tries, "Test couldn't get result it wanted after many tries")
            tries += 1

            self._connect()

            stream_handler = StreamHandler(self._on_handler_freakout)
            operation = self.echo_client.new_echo_stream_messages(stream_handler)
            stream_handler.operation = operation

            operation.activate(model.EchoStreamingRequest())
            close_future = self.connection.close()

            try:
                response = operation.get_response().result(TIMEOUT)
            except StreamClosedError:
                closed_before_response = True

            # wait for close to complete before attempting reconnect
            close_future.result(TIMEOUT)



from awscrt import mqtt5, http
from awscrt.auth import AwsCredentialsProvider
from awscrt.io import ClientBootstrap, DefaultHostResolver, EventLoopGroup
from awsiot import mqtt5_client_builder
from concurrent.futures import Future
import boto3
import botocore.exceptions
import os
import unittest
import shutil
import tempfile
import uuid
import warnings

TIMEOUT = 100.0
PROXY_HOST = os.environ.get('proxyhost')
PROXY_PORT = int(os.environ.get('proxyport', '0'))

CUSTOM_AUTHORIZER_ENDPOINT = os.environ.get("CUSTOM_AUTHORIZER_ENDPOINT")
CUSTOM_AUTHORIZER_NAME_SIGNED = os.environ.get("CUSTOM_AUTHORIZER_NAME_SIGNED")
CUSTOM_AUTHORIZER_NAME_UNSIGNED = os.environ.get("CUSTOM_AUTHORIZER_NAME_UNSIGNED")
CUSTOM_AUTHORIZER_PASSWORD = os.environ.get("CUSTOM_AUTHORIZER_PASSWORD")
CUSTOM_AUTHORIZER_SIGNATURE = os.environ.get("CUSTOM_AUTHORIZER_SIGNATURE")
CUSTOM_AUTHORIZER_SIGNATURE_UNENCODED = os.environ.get("CUSTOM_AUTHORIZER_SIGNATURE_UNENCODED")
CUSTOM_AUTHORIZER_TOKEN_KEY_NAME = os.environ.get("CUSTOM_AUTHORIZER_TOKEN_KEY_NAME")
CUSTOM_AUTHORIZER_TOKEN_VALUE = os.environ.get("CUSTOM_AUTHORIZER_TOKEN_VALUE")


def has_custom_auth_environment():
    return (CUSTOM_AUTHORIZER_ENDPOINT is not None) and (CUSTOM_AUTHORIZER_NAME_SIGNED is not None) and \
           (CUSTOM_AUTHORIZER_NAME_UNSIGNED is not None) and (CUSTOM_AUTHORIZER_PASSWORD is not None) and \
           (CUSTOM_AUTHORIZER_SIGNATURE is not None) and (CUSTOM_AUTHORIZER_TOKEN_KEY_NAME is not None) and \
           (CUSTOM_AUTHORIZER_TOKEN_VALUE is not None) and (CUSTOM_AUTHORIZER_SIGNATURE_UNENCODED is not None)

class Config:
    cache = None

    def __init__(self, endpoint, cert, key, region, cognito_creds):
        self.cert = cert
        self.key = key
        self.endpoint = endpoint
        self.region = region
        self.cognito_creds = cognito_creds

    @staticmethod
    def get():
        """Raises SkipTest if credentials aren't set up correctly"""
        if Config.cache:
            return Config.cache

        # boto3 caches the HTTPS connection for the API calls, which appears to the unit test
        # framework as a leak, so ignore it, that's not what we're testing here
        warnings.simplefilter('ignore', ResourceWarning)

        try:
            secrets = boto3.client('secretsmanager')
            response = secrets.get_secret_value(SecretId='unit-test/endpoint')
            endpoint = response['SecretString']
            response = secrets.get_secret_value(SecretId='unit-test/certificate')
            cert = response['SecretString'].encode('utf8')
            response = secrets.get_secret_value(SecretId='unit-test/privatekey')
            key = response['SecretString'].encode('utf8')
            region = secrets.meta.region_name
            response = secrets.get_secret_value(SecretId='unit-test/cognitopool')
            cognito_pool = response['SecretString']

            cognito = boto3.client('cognito-identity')
            response = cognito.get_id(IdentityPoolId=cognito_pool)
            cognito_id = response['IdentityId']
            response = cognito.get_credentials_for_identity(IdentityId=cognito_id)
            cognito_creds = response['Credentials']

            Config.cache = Config(endpoint, cert, key, region, cognito_creds)
        except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as ex:
            raise unittest.SkipTest("No credentials")

        return Config.cache


def create_client_id():
    return 'aws-iot-device-sdk-python-v2-unit-test-{0}'.format(uuid.uuid4())

class Mqtt5TestCallbacks():
    def __init__(self):
        self.future_connection_success = Future()
        self.future_stopped = Future()

    def ws_handshake_transform(self, transform_args):
        transform_args.set_done()

    def on_lifecycle_stopped(self, lifecycle_stopped: mqtt5.LifecycleStoppedData):
        self.future_stopped.set_result(None)

    def on_lifecycle_connection_success(self, lifecycle_connection_success: mqtt5.LifecycleConnectSuccessData):
        self.future_connection_success.set_result(lifecycle_connection_success)

class Mqtt5BuilderTest(unittest.TestCase):
    def _test_connection(self, client: mqtt5.Client, callbacks: Mqtt5TestCallbacks):
        client.start()
        callbacks.future_connection_success.result(TIMEOUT)
        client.stop()
        callbacks.future_stopped.result(TIMEOUT)

    def test_mtls_from_bytes(self):
        config = Config.get()
        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)
        callbacks = Mqtt5TestCallbacks()

        client = mqtt5_client_builder.mtls_from_bytes(
            cert_bytes=config.cert,
            pri_key_bytes=config.key,
            endpoint=config.endpoint,
            client_id=create_client_id(),
            client_bootstrap=bootstrap,
            on_lifecycle_connection_success=callbacks.on_lifecycle_connection_success,
            on_lifecycle_stopped=callbacks.on_lifecycle_stopped)

        self._test_connection(client, callbacks)

    def test_mtls_from_path(self):
        config = Config.get()
        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)
        callbacks = Mqtt5TestCallbacks()

        # test "from path" builder by writing secrets to tempfiles
        tmp_dirpath = tempfile.mkdtemp()
        try:
            cert_filepath = os.path.join(tmp_dirpath, 'cert')
            with open(cert_filepath, 'wb') as cert_file:
                cert_file.write(config.cert)

            key_filepath = os.path.join(tmp_dirpath, 'key')
            with open(key_filepath, 'wb') as key_file:
                key_file.write(config.key)

            client = mqtt5_client_builder.mtls_from_path(
                cert_filepath=cert_filepath,
                pri_key_filepath=key_filepath,
                endpoint=config.endpoint,
                client_id=create_client_id(),
                client_bootstrap=bootstrap,
                on_lifecycle_connection_success=callbacks.on_lifecycle_connection_success,
                on_lifecycle_stopped=callbacks.on_lifecycle_stopped)

        finally:
            shutil.rmtree(tmp_dirpath)

        self._test_connection(client, callbacks)

    def test_websockets_default(self):
        """Websocket connection with default credentials provider"""
        config = Config.get()
        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)
        cred_provider = AwsCredentialsProvider.new_default_chain(bootstrap)
        callbacks = Mqtt5TestCallbacks()

        client = mqtt5_client_builder.websockets_with_default_aws_signing(
            region=config.region,
            credentials_provider=cred_provider,
            endpoint=config.endpoint,
            client_id=create_client_id(),
            client_bootstrap=bootstrap,
            on_lifecycle_connection_success=callbacks.on_lifecycle_connection_success,
            on_lifecycle_stopped=callbacks.on_lifecycle_stopped)

        self._test_connection(client, callbacks)

    def test_websockets_sts(self):
        """Websocket connection with X-Amz-Security-Token query param"""
        config = Config.get()
        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)
        cred_provider = AwsCredentialsProvider.new_static(
            access_key_id=config.cognito_creds['AccessKeyId'],
            secret_access_key=config.cognito_creds['SecretKey'],
            session_token=config.cognito_creds['SessionToken'])
        callbacks = Mqtt5TestCallbacks()

        client = mqtt5_client_builder.websockets_with_default_aws_signing(
            region=config.region,
            credentials_provider=cred_provider,
            endpoint=config.endpoint,
            client_id=create_client_id(),
            client_bootstrap=bootstrap,
            on_lifecycle_connection_success=callbacks.on_lifecycle_connection_success,
            on_lifecycle_stopped=callbacks.on_lifecycle_stopped)

        self._test_connection(client, callbacks)

    @unittest.skipIf(PROXY_HOST is None, 'requires "proxyhost" and "proxyport" env vars')
    def test_websockets_proxy(self):
        config = Config.get()
        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)
        cred_provider = AwsCredentialsProvider.new_default_chain(bootstrap)
        callbacks = Mqtt5TestCallbacks()

        client = mqtt5_client_builder.websockets_with_default_aws_signing(
            region=config.region,
            credentials_provider=cred_provider,
            endpoint=config.endpoint,
            websocket_proxy_options=http.HttpProxyOptions(PROXY_HOST, PROXY_PORT),
            client_id=create_client_id(),
            client_bootstrap=bootstrap,
            on_lifecycle_connection_success=callbacks.on_lifecycle_connection_success,
            on_lifecycle_stopped=callbacks.on_lifecycle_stopped)

        self._test_connection(client, callbacks)

    @unittest.skipIf(not has_custom_auth_environment(), 'requires custom authentication env vars')
    def test_mqtt5_builder_direct_signed_custom_authorizer(self):
        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)
        callbacks = Mqtt5TestCallbacks()

        client = mqtt5_client_builder.direct_with_custom_authorizer(
            auth_username="",
            auth_authorizer_name=CUSTOM_AUTHORIZER_NAME_SIGNED,
            auth_authorizer_signature=CUSTOM_AUTHORIZER_SIGNATURE,
            auth_password=CUSTOM_AUTHORIZER_PASSWORD,
            auth_token_key_name=CUSTOM_AUTHORIZER_TOKEN_KEY_NAME,
            auth_token_value=CUSTOM_AUTHORIZER_TOKEN_VALUE,
            endpoint=CUSTOM_AUTHORIZER_ENDPOINT,
            client_id=create_client_id(),
            client_bootstrap=bootstrap,
            on_lifecycle_connection_success=callbacks.on_lifecycle_connection_success,
            on_lifecycle_stopped=callbacks.on_lifecycle_stopped)

        self._test_connection(client, callbacks)

    @unittest.skipIf(not has_custom_auth_environment(), 'requires custom authentication env vars')
    def test_mqtt5_builder_direct_signed_custom_authorizer_unencoded(self):
        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)
        callbacks = Mqtt5TestCallbacks()

        client = mqtt5_client_builder.direct_with_custom_authorizer(
            auth_username="",
            auth_authorizer_name=CUSTOM_AUTHORIZER_NAME_SIGNED,
            auth_authorizer_signature=CUSTOM_AUTHORIZER_SIGNATURE_UNENCODED,
            auth_password=CUSTOM_AUTHORIZER_PASSWORD,
            auth_token_key_name=CUSTOM_AUTHORIZER_TOKEN_KEY_NAME,
            auth_token_value=CUSTOM_AUTHORIZER_TOKEN_VALUE,
            endpoint=CUSTOM_AUTHORIZER_ENDPOINT,
            client_id=create_client_id(),
            client_bootstrap=bootstrap,
            on_lifecycle_connection_success=callbacks.on_lifecycle_connection_success,
            on_lifecycle_stopped=callbacks.on_lifecycle_stopped)

        self._test_connection(client, callbacks)

    @unittest.skipIf(not has_custom_auth_environment(), 'requires custom authentication env vars')
    def test_mqtt5_builder_direct_unsigned_custom_authorizer(self):
        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)
        callbacks = Mqtt5TestCallbacks()

        client = mqtt5_client_builder.direct_with_custom_authorizer(
            auth_username="",
            auth_authorizer_name=CUSTOM_AUTHORIZER_NAME_UNSIGNED,
            auth_password=CUSTOM_AUTHORIZER_PASSWORD,
            endpoint=CUSTOM_AUTHORIZER_ENDPOINT,
            client_id=create_client_id(),
            client_bootstrap=bootstrap,
            on_lifecycle_connection_success=callbacks.on_lifecycle_connection_success,
            on_lifecycle_stopped=callbacks.on_lifecycle_stopped)

        self._test_connection(client, callbacks)

    @unittest.skipIf(not has_custom_auth_environment(), 'requires custom authentication env vars')
    def test_mqtt5_builder_websocket_signed_custom_authorizer(self):
        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)
        callbacks = Mqtt5TestCallbacks()

        client = mqtt5_client_builder.websockets_with_custom_authorizer(
            auth_username="",
            auth_authorizer_name=CUSTOM_AUTHORIZER_NAME_SIGNED,
            auth_authorizer_signature=CUSTOM_AUTHORIZER_SIGNATURE,
            auth_password=CUSTOM_AUTHORIZER_PASSWORD,
            auth_token_key_name=CUSTOM_AUTHORIZER_TOKEN_KEY_NAME,
            auth_token_value=CUSTOM_AUTHORIZER_TOKEN_VALUE,
            endpoint=CUSTOM_AUTHORIZER_ENDPOINT,
            client_id=create_client_id(),
            client_bootstrap=bootstrap,
            on_lifecycle_connection_success=callbacks.on_lifecycle_connection_success,
            on_lifecycle_stopped=callbacks.on_lifecycle_stopped)

        self._test_connection(client, callbacks)

    @unittest.skipIf(not has_custom_auth_environment(), 'requires custom authentication env vars')
    def test_mqtt5_builder_websocket_signed_custom_authorizer_unencoded(self):
        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)
        callbacks = Mqtt5TestCallbacks()

        client = mqtt5_client_builder.websockets_with_custom_authorizer(
            auth_username="",
            auth_authorizer_name=CUSTOM_AUTHORIZER_NAME_SIGNED,
            auth_authorizer_signature=CUSTOM_AUTHORIZER_SIGNATURE_UNENCODED,
            auth_password=CUSTOM_AUTHORIZER_PASSWORD,
            auth_token_key_name=CUSTOM_AUTHORIZER_TOKEN_KEY_NAME,
            auth_token_value=CUSTOM_AUTHORIZER_TOKEN_VALUE,
            endpoint=CUSTOM_AUTHORIZER_ENDPOINT,
            client_id=create_client_id(),
            client_bootstrap=bootstrap,
            on_lifecycle_connection_success=callbacks.on_lifecycle_connection_success,
            on_lifecycle_stopped=callbacks.on_lifecycle_stopped)

        self._test_connection(client, callbacks)

    @unittest.skipIf(not has_custom_auth_environment(), 'requires custom authentication env vars')
    def test_mqtt5_builder_websocket_unsigned_custom_authorizer(self):
        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)
        callbacks = Mqtt5TestCallbacks()

        client = mqtt5_client_builder.websockets_with_custom_authorizer(
            auth_username="",
            auth_authorizer_name=CUSTOM_AUTHORIZER_NAME_UNSIGNED,
            auth_password=CUSTOM_AUTHORIZER_PASSWORD,
            endpoint=CUSTOM_AUTHORIZER_ENDPOINT,
            client_id=create_client_id(),
            client_bootstrap=bootstrap,
            on_lifecycle_connection_success=callbacks.on_lifecycle_connection_success,
            on_lifecycle_stopped=callbacks.on_lifecycle_stopped)

        self._test_connection(client, callbacks)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0.

from awscrt.http import HttpClientConnection, HttpRequest, HttpHeaders, HttpProxyOptions
from awscrt.io import ClientBootstrap, ClientTlsContext, is_alpn_available, SocketOptions, TlsConnectionOptions
import awsiot
from concurrent.futures import Future
import json
from typing import Any, Dict, List, Optional


class DiscoveryClient:
    """
    Client which performs Greengrass discovery.

    Args:
        bootstrap: Client bootstrap
        socket_options: Socket options
        tls_context: Client TLS context
        region: AWS region (not used if gg_server_name is set)
        gg_server_name: optional full server name
        proxy_options: Proxy options (if None is provided then a proxy is not used)
    """
    __slots__ = [
        '_bootstrap',
        '_tls_context',
        '_socket_options',
        '_region',
        '_tls_connection_options',
        '_gg_server_name',
        'gg_url',
        'port',
        "_proxy_options"]

    def __init__(
            self,
            bootstrap: ClientBootstrap,
            socket_options: SocketOptions,
            tls_context: ClientTlsContext,
            region: str,
            gg_server_name: str = None,
            proxy_options: HttpProxyOptions = None):
        assert isinstance(bootstrap, ClientBootstrap)
        assert isinstance(socket_options, SocketOptions)
        assert isinstance(tls_context, ClientTlsContext)
        assert isinstance(region, str)
        if gg_server_name is not None:
            assert isinstance(gg_server_name, str)

        self._bootstrap = bootstrap
        self._socket_options = socket_options
        self._region = region
        if gg_server_name is None:
            self._gg_server_name = 'greengrass-ats.iot.{}.amazonaws.com'.format(region)
        else:
            self._gg_server_name = gg_server_name

        self._tls_connection_options = tls_context.new_connection_options()
        self._tls_connection_options.set_server_name(self._gg_server_name)
        self.port = 8443
        self._proxy_options = proxy_options

        if is_alpn_available():
            self._tls_connection_options.set_alpn_list(['x-amzn-http-ca'])
            self.port = 443

    def discover(self, thing_name: str) -> Future:
        """
        Perform discovery.

        This is an asynchronous operation.

        Returns:
            Future a Future which will contain a result of :class:`DiscoverResponse`
            on success, or an exception on failure.
        """

        discovery = dict(
            future=Future(),
            response_body=bytearray())

        def on_incoming_body(http_stream, chunk, **kwargs):
            discovery['response_body'].extend(chunk)

        def on_request_complete(completion_future):
            try:
                response_code = completion_future.result()
                if response_code == 200:
                    payload_str = discovery['response_body'].decode('utf-8')
                    discover_res = DiscoverResponse.from_payload(json.loads(payload_str))
                    discovery['future'].set_result(discover_res)
                else:
                    discovery['future'].set_exception(
                        DiscoveryException(
                            'Error during discover call: response_code={}'.format(response_code),
                            response_code))

            except Exception as e:
                discovery['future'].set_exception(e)

        def on_connection_completed(conn_future):
            try:
                connection = conn_future.result()
                headers = HttpHeaders()
                headers.add('host', self._gg_server_name)
                request = HttpRequest(
                    method='GET',
                    path='/greengrass/discover/thing/{}'.format(thing_name),
                    headers=headers)

                http_stream = connection.request(
                    request=request,
                    on_body=on_incoming_body)

                http_stream.activate()
                http_stream.completion_future.add_done_callback(on_request_complete)

            except Exception as e:
                discovery['future'].set_exception(e)

        connect_future = HttpClientConnection.new(
            host_name=self._gg_server_name,
            port=self.port,
            socket_options=self._socket_options,
            tls_connection_options=self._tls_connection_options,
            bootstrap=self._bootstrap,
            proxy_options=self._proxy_options)

        connect_future.add_done_callback(on_connection_completed)

        return discovery['future']


class DiscoveryException(Exception):
    """
    Discovery response was an error.
    """
    _slots_ = ['http_response_code', 'message']

    def __init__(self, message: str, response_code:int):
        #: HTTP response code
        self.http_response_code = response_code # type: int
        #: Message
        self.message = message # type: str


class ConnectivityInfo(awsiot.ModeledClass):
    """
    Connectivity info
    """

    __slots__ = ['id', 'host_address', 'metadata', 'port']
    def __init__(self):
        #: ID
        self.id = None
        #: Port address
        self.host_address = None
        #: Metadata
        self.metadata = None
        #: Port
        self.port = None

    @classmethod
    def from_payload(cls, payload: Dict[str, Any]) -> 'ConnectivityInfo':
        new = cls()
        val = payload.get('Id')
        if val is not None:
            new.id = val
        val = payload.get('HostAddress')
        if val is not None:
            new.host_address = val
        val = payload.get('PortNumber')
        if val is not None:
            new.port = val
        val = payload.get('Metadata')
        if val is not None:
            new.metadata = val
        return new


class GGCore(awsiot.ModeledClass):
    """
    Greengrass Core
    """
    __slots__ = ['thing_arn', 'connectivity']

    def __init__(self):
        #: Thing ARN
        self.thing_arn = None
        #: List of :class:`ConnectivityInfo`
        self.connectivity = None

    @classmethod
    def from_payload(cls, payload: Dict[str, Any]) -> 'GGCore':
        new = cls()
        val = payload.get('thingArn')
        if val is not None:
            new.thing_arn = val
        val = payload.get('Connectivity')
        if val is not None:
            new.connectivity = [ConnectivityInfo.from_payload(i) for i in val]

        return new


class GGGroup(awsiot.ModeledClass):
    """
    Greengrass group
    """
    __slots__ = ['gg_group_id', 'cores', 'certificate_authorities']

    def __init__(self):
        #: Greengrass group ID
        self.gg_group_id = None
        #: List of :class:`GGCore`
        self.cores = None
        #: List of strings
        self.certificate_authorities = None

    @classmethod
    def from_payload(cls, payload: Dict[str, Any]) -> 'GGGroup':
        new = cls()
        val = payload.get('GGGroupId')
        if val is not None:
            new.gg_group_id = val
        val = payload.get('Cores')
        if val is not None:
            new.cores = [GGCore.from_payload(i) for i in val]
        val = payload.get('CAs')
        if val is not None:
            new.certificate_authorities = val

        return new


class DiscoverResponse(awsiot.ModeledClass):
    """
    Discovery response
    """
    __slots__ = ['gg_groups']

    def __init__(self):
        #: List of :class:`GGGroup`
        self.gg_groups = None # type: Optional[List[GGGroup]]

    @classmethod
    def from_payload(cls, payload: Dict[str, Any]) -> 'DiscoverResponse':
        new = cls()
        val = payload.get('GGGroups')
        if val is not None:
            new.gg_groups = [GGGroup.from_payload(i) for i in val]

        return new



"""
Builder functions to create a :class:`awscrt.mqtt.Connection`, configured for use with AWS IoT Core.
The following keyword arguments are common to all builder functions:

Required Keyword Arguments:

    **endpoint** (`str`): Host name of AWS IoT server.

    **client_id** (`str`): ID to place in CONNECT packet. Must be unique across all devices/clients.
            If an ID is already in use, the other client will be disconnected.

Optional Keyword Arguments (omit, or set `None` to get default value):

    **client_bootstrap** (:class:`awscrt.io.ClientBootstrap`): Client bootstrap used to establish connection.
        The ClientBootstrap will default to the static default (Io.ClientBootstrap.get_or_create_static_default)
        if the argument is omitted or set to 'None'.

    **on_connection_interrupted** (`Callable`): Callback invoked whenever the MQTT connection is lost.
        The MQTT client will automatically attempt to reconnect.
        The function should take the following arguments return nothing:

            *   `connection` (:class:`awscrt.mqtt.Connection`): This MQTT Connection.

            *   `error` (:class:`awscrt.exceptions.AwsCrtError`): Exception which caused connection loss.

            *   `**kwargs` (dict): Forward-compatibility kwargs.

    **on_connection_resumed** (`Callable`): Callback invoked whenever the MQTT connection
        is automatically resumed. Function should take the following arguments and return nothing:

            *   `connection` (:class:`awscrt.mqtt.Connection`): This MQTT Connection

            *   `return_code` (:class:`awscrt.mqtt.ConnectReturnCode`): Connect return
                code received from the server.

            *   `session_present` (bool): True if resuming existing session. False if new session.
                Note that the server has forgotten all previous subscriptions if this is False.
                Subscriptions can be re-established via resubscribe_existing_topics().

            *   `**kwargs` (dict): Forward-compatibility kwargs.

    **on_connection_success** (`Callable`): Optional callback invoked whenever the connection successfully connects.
        The function should take the following arguments and return nothing:

            *   `connection` (:class:`awscrt.mqtt.Connection`): This MQTT Connection.

            *   `callback_data` (:class:`awscrt.mqtt.OnConnectionSuccessData`): The data returned from the connection success.

    **on_connection_failure** (`Callable`): Optional callback invoked whenever the connection fails to connect.
        The function should take the following arguments and return nothing:

            *   `connection` (:class:`awscrt.mqtt.Connection`): This MQTT Connection.

            *   `callback_data` (:class:`awscrt.mqtt.OnConnectionFailureData`): The data returned from the connection failure.

    **on_connection_closed** (`Callable`): Optional callback invoked whenever the connection has been disconnected and shutdown successfully.
        The function should take the following arguments and return nothing:

            *   `connection` (:class:`awscrt.mqtt.Connection`): This MQTT Connection.

            *   `callback_data` (:class:`awscrt.mqtt.OnConnectionClosedData`): The data returned from the connection close.

    **clean_session** (`bool`): Whether or not to start a clean session with each reconnect.
        If True, the server will forget all subscriptions with each reconnect.
        Set False to request that the server resume an existing session
        or start a new session that may be resumed after a connection loss.
        The `session_present` bool in the connection callback informs
        whether an existing session was successfully resumed.
        If an existing session is resumed, the server remembers previous subscriptions
        and sends mesages (with QoS1 or higher) that were published while the client was offline.

    **reconnect_min_timeout_secs** (`int`): Minimum time to wait between reconnect attempts.
        Must be <= `reconnect_max_timeout_secs`.
        Wait starts at min and doubles with each attempt until max is reached.

    **reconnect_max_timeout_secs** (`int`): Maximum time to wait between reconnect attempts.
        Must be >= `reconnect_min_timeout_secs`.
        Wait starts at min and doubles with each attempt until max is reached.

    **keep_alive_secs** (`int`): The keep alive value, in seconds, to send in CONNECT packet.
        A PING will automatically be sent at this interval.
        The server will assume the connection is lost if no PING is received after 1.5X this value.
        Default is 1200sec (20 minutes). This duration must be longer than ping_timeout_ms.

    **ping_timeout_ms** (`int`): Milliseconds to wait for ping response before client assumes
        the connection is invalid and attempts to reconnect.
        Default is 3000ms (3 seconds). This duration must be shorter than `keep_alive_secs`.

    **protocol_operation_timeout_ms** (`int`): Milliseconds to wait for the response to the operation
        requires response by protocol. Set to zero to disable timeout. Otherwise,
        the operation will fail if no response is received within this amount of time after
        the packet is written to the socket
        It applied to PUBLISH (QoS>0) and UNSUBSCRIBE now.

    **will** (:class:`awscrt.mqtt.Will`): Will to send with CONNECT packet. The will is
        published by the server when its connection to the client is unexpectedly lost.

    **username** (`str`): Username to connect with.

    **password** (`str`): Password to connect with.

    **port** (`int`): Override default server port.
        Default port is 443 if system supports ALPN or websockets are being used.
        Otherwise, default port is 8883.

    **tcp_connect_timeout_ms** (`int`): Milliseconds to wait for TCP connect response. Default is 5000ms (5 seconds).

    **ca_filepath** (`str`): Override default trust store with CA certificates from this PEM formatted file.

    **ca_dirpath** (`str`): Override default trust store with CA certificates loaded from this directory (Unix only).

    **ca_bytes** (`bytes`): Override default trust store with CA certificates from these PEM formatted bytes.

    **enable_metrics_collection** (`bool`): Whether to send the SDK version number in the CONNECT packet.
        Default is True.

    **http_proxy_options** (:class: 'awscrt.http.HttpProxyOptions'): HTTP proxy options to use
"""

# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0.

import awscrt.auth
import awscrt.io
import awscrt.mqtt
import urllib.parse


def _check_required_kwargs(**kwargs):
    for required in ['endpoint', 'client_id']:
        if not kwargs.get(required):
            raise TypeError("Builder needs keyword-only argument '{}'".format(required))


def _get(kwargs, name, default=None):
    """
    Returns kwargs['name'] if it exists and is not None.
    Otherwise returns default.

    This function exists so users can pass some_arg=None to get its default
    value, instead of literally passing None.
    """
    val = kwargs.get(name)
    if val is None:
        val = default
    return val


_metrics_str = None


def _get_metrics_str(current_username=""):
    global _metrics_str

    username_has_query = False
    if current_username.find("?") != -1:
        username_has_query = True

    if _metrics_str is None:
        try:
            import pkg_resources
            try:
                version = pkg_resources.get_distribution("awsiotsdk").version
                _metrics_str = "SDK=PythonV2&Version={}".format(version)
            except pkg_resources.DistributionNotFound:
                _metrics_str = "SDK=PythonV2&Version=dev"
        except BaseException:
            _metrics_str = ""

    if not _metrics_str == "":
        if username_has_query:
            return "&" + _metrics_str
        else:
            return "?" + _metrics_str
    else:
        return ""


def _builder(
        tls_ctx_options,
        use_websockets=False,
        websocket_handshake_transform=None,
        use_custom_authorizer=False,
        **kwargs):

    ca_bytes = _get(kwargs, 'ca_bytes')
    ca_filepath = _get(kwargs, 'ca_filepath')
    ca_dirpath = _get(kwargs, 'ca_dirpath')
    if ca_bytes:
        tls_ctx_options.override_default_trust_store(ca_bytes)
    elif ca_filepath or ca_dirpath:
        tls_ctx_options.override_default_trust_store_from_path(ca_dirpath, ca_filepath)

    port = _get(kwargs, 'port')
    if port is None:
        # prefer 443, even for direct MQTT connections, since it's less likely to be blocked by firewalls
        if use_websockets or awscrt.io.is_alpn_available():
            port = 443
        else:
            port = 8883

    if port == 443 and awscrt.io.is_alpn_available() and use_custom_authorizer is False:
        tls_ctx_options.alpn_list = ['http/1.1'] if use_websockets else ['x-amzn-mqtt-ca']

    socket_options = awscrt.io.SocketOptions()
    socket_options.connect_timeout_ms = _get(kwargs, 'tcp_connect_timeout_ms', 5000)
    # These have been inconsistent between keepalive/keep_alive. Resolve both for now to ease transition.
    socket_options.keep_alive = \
        _get(kwargs, 'tcp_keep_alive', _get(kwargs, 'tcp_keepalive', False))

    socket_options.keep_alive_timeout_secs = \
        _get(kwargs, 'tcp_keep_alive_timeout_secs', _get(kwargs, 'tcp_keepalive_timeout_secs', 0))

    socket_options.keep_alive_interval_secs = \
        _get(kwargs, 'tcp_keep_alive_interval_secs', _get(kwargs, 'tcp_keepalive_interval_secs', 0))

    socket_options.keep_alive_max_probes = \
        _get(kwargs, 'tcp_keep_alive_max_probes', _get(kwargs, 'tcp_keepalive_max_probes', 0))

    username = _get(kwargs, 'username', '')
    if _get(kwargs, 'enable_metrics_collection', True):
        username += _get_metrics_str(username)

    if username == "":
        username = None

    client_bootstrap = _get(kwargs, 'client_bootstrap')
    if client_bootstrap is None:
        client_bootstrap = awscrt.io.ClientBootstrap.get_or_create_static_default()

    tls_ctx = awscrt.io.ClientTlsContext(tls_ctx_options)
    mqtt_client = awscrt.mqtt.Client(client_bootstrap, tls_ctx)

    proxy_options = kwargs.get('http_proxy_options', kwargs.get('websocket_proxy_options', None))
    return awscrt.mqtt.Connection(
        client=mqtt_client,
        on_connection_interrupted=_get(kwargs, 'on_connection_interrupted'),
        on_connection_resumed=_get(kwargs, 'on_connection_resumed'),
        client_id=_get(kwargs, 'client_id'),
        host_name=_get(kwargs, 'endpoint'),
        port=port,
        clean_session=_get(kwargs, 'clean_session', False),
        reconnect_min_timeout_secs=_get(kwargs, 'reconnect_min_timeout_secs', 5),
        reconnect_max_timeout_secs=_get(kwargs, 'reconnect_max_timeout_secs', 60),
        keep_alive_secs=_get(kwargs, 'keep_alive_secs', 1200),
        ping_timeout_ms=_get(kwargs, 'ping_timeout_ms', 3000),
        protocol_operation_timeout_ms=_get(kwargs, 'protocol_operation_timeout_ms', 0),
        will=_get(kwargs, 'will'),
        username=username,
        password=_get(kwargs, 'password'),
        socket_options=socket_options,
        use_websockets=use_websockets,
        websocket_handshake_transform=websocket_handshake_transform,
        proxy_options=proxy_options,
        on_connection_success=_get(kwargs, 'on_connection_success'),
        on_connection_failure=_get(kwargs, 'on_connection_failure'),
        on_connection_closed=_get(kwargs, 'on_connection_closed'),
    )


def mtls_from_path(cert_filepath, pri_key_filepath, **kwargs) -> awscrt.mqtt.Connection:
    """
    This builder creates an :class:`awscrt.mqtt.Connection`, configured for an mTLS MQTT connection to AWS IoT.
    TLS arguments are passed as filepaths.

    This function takes all :mod:`common arguments<awsiot.mqtt_connection_builder>`
    described at the top of this doc, as well as...

    Keyword Args:
        cert_filepath (str): Path to certificate file.

        pri_key_filepath (str): Path to private key file.
    """
    _check_required_kwargs(**kwargs)
    tls_ctx_options = awscrt.io.TlsContextOptions.create_client_with_mtls_from_path(cert_filepath, pri_key_filepath)
    return _builder(tls_ctx_options, **kwargs)


def mtls_from_bytes(cert_bytes, pri_key_bytes, **kwargs) -> awscrt.mqtt.Connection:
    """
    This builder creates an :class:`awscrt.mqtt.Connection`, configured for an mTLS MQTT connection to AWS IoT.
    TLS arguments are passed as in-memory bytes.

    This function takes all :mod:`common arguments<awsiot.mqtt_connection_builder>`
    described at the top of this doc, as well as...

    Keyword Args:
        cert_bytes (bytes): Certificate file bytes.

        pri_key_bytes (bytes): Private key bytes.
    """
    _check_required_kwargs(**kwargs)
    tls_ctx_options = awscrt.io.TlsContextOptions.create_client_with_mtls(cert_bytes, pri_key_bytes)
    return _builder(tls_ctx_options, **kwargs)


def mtls_with_pkcs11(*,
                     pkcs11_lib: awscrt.io.Pkcs11Lib,
                     user_pin: str,
                     slot_id: int = None,
                     token_label: str = None,
                     private_key_label: str = None,
                     cert_filepath: str = None,
                     cert_bytes=None,
                     **kwargs) -> awscrt.mqtt.Connection:
    """
    This builder creates an :class:`awscrt.mqtt.Connection`, configured for an mTLS MQTT connection to AWS IoT,
    using a PKCS#11 library for private key operations.

    NOTE: Unix only

    This function takes all :mod:`common arguments<awsiot.mqtt_connection_builder>`
    described at the top of this doc, as well as...

    Args:
        pkcs11_lib: Use this PKCS#11 library

        user_pin: User PIN, for logging into the PKCS#11 token.
            Pass `None` to log into a token with a "protected authentication path".

        slot_id: ID of slot containing PKCS#11 token.
            If not specified, the token will be chosen based on other criteria (such as token label).

        token_label: Label of the PKCS#11 token to use.
            If not specified, the token will be chosen based on other criteria (such as slot ID).

        private_key_label: Label of private key object on PKCS#11 token.
            If not specified, the key will be chosen based on other criteria
            (such as being the only available private key on the token).

        cert_filepath: Use this X.509 certificate (file on disk).
            The certificate must be PEM-formatted. The certificate may be
            specified by other means instead (ex: `cert_bytes`)

        cert_bytes (Optional[Union[str, bytes, bytearray]]):
            Use this X.509 certificate (contents in memory).
            The certificate must be PEM-formatted. The certificate may be
            specified by other means instead (ex: `cert_filepath`)
    """
    _check_required_kwargs(**kwargs)

    tls_ctx_options = awscrt.io.TlsContextOptions.create_client_with_mtls_pkcs11(
        pkcs11_lib=pkcs11_lib,
        user_pin=user_pin,
        slot_id=slot_id,
        token_label=token_label,
        private_key_label=private_key_label,
        cert_file_path=cert_filepath,
        cert_file_contents=cert_bytes)

    return _builder(tls_ctx_options, **kwargs)

def mtls_with_pkcs12(*,
                     pkcs12_filepath: str,
                     pkcs12_password: str,
                     **kwargs) -> awscrt.mqtt.Connection:
    """
    This builder creates an :class:`awscrt.mqtt.Connection`, configured for an mTLS MQTT connection to AWS IoT,
    using a PKCS#12 certificate.

    NOTE: MacOS only

    This function takes all :mod:`common arguments<awsiot.mqtt_connection_builder>`
    described at the top of this doc, as well as...

    Args:
        pkcs12_filepath: Path to the PKCS12 file to use

        pkcs12_password: The password for the PKCS12 file.
    """
    _check_required_kwargs(**kwargs)

    tls_ctx_options = awscrt.io.TlsContextOptions.create_client_with_mtls_pkcs12(
        pkcs12_filepath=pkcs12_filepath,
        pkcs12_password=pkcs12_password)
    return _builder(tls_ctx_options, **kwargs)


def mtls_with_windows_cert_store_path(*,
                                      cert_store_path: str,
                                      **kwargs) -> awscrt.mqtt.Connection:
    """
    This builder creates an :class:`awscrt.mqtt.Connection`, configured for an mTLS MQTT connection to AWS IoT,
    using a client certificate in a Windows certificate store.

    NOTE: Windows only

    This function takes all :mod:`common arguments<awsiot.mqtt_connection_builder>`
    described at the top of this doc, as well as...

    Args:
        cert_store_path: Path to certificate in a Windows certificate store.
                The path must use backslashes and end with the certificate's thumbprint.
                Example: ``CurrentUser\\MY\\A11F8A9B5DF5B98BA3508FBCA575D09570E0D2C6``
    """
    _check_required_kwargs(**kwargs)

    tls_ctx_options = awscrt.io.TlsContextOptions.create_client_with_mtls_windows_cert_store_path(cert_store_path)

    return _builder(tls_ctx_options, **kwargs)


def websockets_with_default_aws_signing(
        region,
        credentials_provider,
        websocket_proxy_options=None,
        **kwargs) -> awscrt.mqtt.Connection:
    """
    This builder creates an :class:`awscrt.mqtt.Connection`, configured for an MQTT connection over websockets to AWS IoT.
    The websocket handshake is signed using credentials from the credentials_provider.

    This function takes all :mod:`common arguments<awsiot.mqtt_connection_builder>`
    described at the top of this doc, as well as...

    Keyword Args:
        region (str): AWS region to use when signing.

        credentials_provider (awscrt.auth.AwsCredentialsProvider): Source of AWS credentials to use when signing.

        websocket_proxy_options (awscrt.http.HttpProxyOptions): Deprecated,
            for proxy settings use `http_proxy_options` (described in
            :mod:`common arguments<awsiot.mqtt_connection_builder>`)

    """
    _check_required_kwargs(**kwargs)

    def _sign_websocket_handshake_request(transform_args, **kwargs):
        # transform_args need to know when transform is done
        try:
            signing_config = awscrt.auth.AwsSigningConfig(
                algorithm=awscrt.auth.AwsSigningAlgorithm.V4,
                signature_type=awscrt.auth.AwsSignatureType.HTTP_REQUEST_QUERY_PARAMS,
                credentials_provider=credentials_provider,
                region=region,
                service='iotdevicegateway',
                omit_session_token=True,  # IoT is weird and does not sign X-Amz-Security-Token
            )

            signing_future = awscrt.auth.aws_sign_request(transform_args.http_request, signing_config)
            signing_future.add_done_callback(lambda x: transform_args.set_done(x.exception()))
        except Exception as e:
            transform_args.set_done(e)

    return websockets_with_custom_handshake(_sign_websocket_handshake_request, websocket_proxy_options, **kwargs)


def websockets_with_custom_handshake(
        websocket_handshake_transform,
        websocket_proxy_options=None,
        **kwargs) -> awscrt.mqtt.Connection:
    """
    This builder creates an :class:`awscrt.mqtt.Connection`, configured for an MQTT connection over websockets,
    with a custom function to transform the websocket handshake request before it is sent to the server.

    This function takes all :mod:`common arguments<awsiot.mqtt_connection_builder>`
    described at the top of this doc, as well as...

    Keyword Args:
        websocket_handshake_transform (Callable): Function to transform websocket handshake request.
            If provided, function is called each time a websocket connection is attempted.
            The function may modify the HTTP request before it is sent to the server.
            See :class:`awscrt.mqtt.WebsocketHandshakeTransformArgs` for more info.
            Function should take the following arguments and return nothing:

                *   `transform_args` (:class:`awscrt.mqtt.WebsocketHandshakeTransformArgs`):
                    Contains HTTP request to be transformed. Function must call
                    `transform_args.done()` when complete.

                *   `**kwargs` (dict): Forward-compatibility kwargs.

        websocket_proxy_options (awscrt.http.HttpProxyOptions):  Deprecated,
            for proxy settings use `http_proxy_options` (described in
            :mod:`common arguments<awsiot.mqtt_connection_builder>`)
    """
    _check_required_kwargs(**kwargs)
    tls_ctx_options = awscrt.io.TlsContextOptions()
    return _builder(tls_ctx_options=tls_ctx_options,
                    use_websockets=True,
                    websocket_handshake_transform=websocket_handshake_transform,
                    websocket_proxy_options=websocket_proxy_options,
                    **kwargs)


def _add_to_username_parameter(input_string, parameter_value, parameter_pretext):
    """
    Helper function to add parameters to the username in the direct_with_custom_authorizer function
    """
    return_string = input_string

    if return_string.find("?") != -1:
        return_string += "&"
    else:
        return_string += "?"

    if parameter_value.find(parameter_pretext) != -1:
        return return_string + parameter_value
    else:
        return return_string + parameter_pretext + parameter_value


def direct_with_custom_authorizer(
        auth_username=None,
        auth_authorizer_name=None,
        auth_authorizer_signature=None,
        auth_password=None,
        auth_token_key_name=None,
        auth_token_value=None,
        **kwargs) -> awscrt.mqtt.Connection:
    """
    This builder creates an :class:`awscrt.mqtt.Connection`, configured for an MQTT connection using a custom
    authorizer using a direct MQTT connection. This function will set the username, port, and TLS options.

    This function takes all :mod:`common arguments<awsiot.mqtt_connection_builder>`
    described at the top of this doc, as well as...

    Keyword Args:
        auth_username (`str`): The username to use with the custom authorizer.
            If provided, the username given will be passed when connecting to the custom authorizer.
            If not provided, it will check to see if a username has already been set (via username="example")
            and will use that instead.  Custom authentication parameters will be appended as appropriate
            to any supplied username value.

        auth_password (`str`):  The password to use with the custom authorizer.
            If not provided, then no password will be sent in the initial CONNECT packet.

        auth_authorizer_name (`str`):  Name of the custom authorizer to use.
            Required if the endpoint does not have a default custom authorizer associated with it.  It is strongly
            suggested to URL-encode this value; the SDK will not do so for you.

        auth_authorizer_signature (`str`):  The digital signature of the token value in the `auth_token_value`
            parameter. The signature must be based on the private key associated with the custom authorizer.  The
            signature must be base64 encoded.
            Required if the custom authorizer has signing enabled.

        auth_token_key_name (`str`): Key used to extract the custom authorizer token from MQTT username query-string
            properties.
            Required if the custom authorizer has signing enabled.  It is strongly suggested to URL-encode
            this value; the SDK will not do so for you.

        auth_token_value (`str`): An opaque token value. This value must be signed by the private key associated with
            the custom authorizer and the result passed in via the `auth_authorizer_signature` parameter.
            Required if the custom authorizer has signing enabled.
    """

    return _with_custom_authorizer(
        auth_username=auth_username,
        auth_authorizer_name=auth_authorizer_name,
        auth_authorizer_signature=auth_authorizer_signature,
        auth_password=auth_password,
        auth_token_key_name=auth_token_key_name,
        auth_token_value=auth_token_value,
        use_websockets=False,
        **kwargs)

def websockets_with_custom_authorizer(
        region=None,
        credentials_provider=None,
        auth_username=None,
        auth_authorizer_name=None,
        auth_authorizer_signature=None,
        auth_password=None,
        auth_token_key_name=None,
        auth_token_value=None,
        **kwargs) -> awscrt.mqtt.Connection:
    """
    This builder creates an :class:`awscrt.mqtt.Connection`, configured for an MQTT connection using a custom
    authorizer using websockets. This function will set the username, port, and TLS options.

    This function takes all :mod:`common arguments<awsiot.mqtt_connection_builder>`
    described at the top of this doc, as well as...

    Keyword Args:
        region (str): AWS region to use when signing.

        credentials_provider (awscrt.auth.AwsCredentialsProvider): Source of AWS credentials to use when signing.

        auth_username (`str`): The username to use with the custom authorizer.
            If provided, the username given will be passed when connecting to the custom authorizer.
            If not provided, it will check to see if a username has already been set (via username="example")
            and will use that instead.  Custom authentication parameters will be appended as appropriate
            to any supplied username value.

        auth_password (`str`):  The password to use with the custom authorizer.
            If not provided, then no password will be sent in the initial CONNECT packet.

        auth_authorizer_name (`str`):  Name of the custom authorizer to use.
            Required if the endpoint does not have a default custom authorizer associated with it.  It is strongly
            suggested to URL-encode this value; the SDK will not do so for you.

        auth_authorizer_signature (`str`):  The digital signature of the token value in the `auth_token_value`
            parameter. The signature must be based on the private key associated with the custom authorizer.  The
            signature must be base64 encoded.
            Required if the custom authorizer has signing enabled.  

        auth_token_key_name (`str`): Key used to extract the custom authorizer token from MQTT username query-string
            properties.
            Required if the custom authorizer has signing enabled.  It is strongly suggested to URL-encode
            this value; the SDK will not do so for you.

        auth_token_value (`str`): An opaque token value. This value must be signed by the private key associated with
            the custom authorizer and the result passed in via the `auth_authorizer_signature` parameter.
            Required if the custom authorizer has signing enabled.
    """

    return _with_custom_authorizer(
        auth_username=auth_username,
        auth_authorizer_name=auth_authorizer_name,
        auth_authorizer_signature=auth_authorizer_signature,
        auth_password=auth_password,
        auth_token_key_name=auth_token_key_name,
        auth_token_value=auth_token_value,
        use_websockets=True,
        websockets_region=region,
        websockets_credentials_provider=credentials_provider,
        **kwargs)


def _with_custom_authorizer(auth_username=None,
        auth_authorizer_name=None,
        auth_authorizer_signature=None,
        auth_password=None,
        auth_token_key_name=None,
        auth_token_value=None,
        use_websockets=False,
        websockets_credentials_provider=None,
        websockets_region=None,
        **kwargs) -> awscrt.mqtt.Connection:
    """
    Helper function that contains the setup needed for custom authorizers
    """

    _check_required_kwargs(**kwargs)
    username_string = ""

    if auth_username is None:
        if not _get(kwargs, "username") is None:
            username_string += _get(kwargs, "username")
    else:
        username_string += auth_username

    if auth_authorizer_name is not None:
        username_string = _add_to_username_parameter(
            username_string, auth_authorizer_name, "x-amz-customauthorizer-name=")

    if auth_authorizer_signature is not None:
        encoded_signature = auth_authorizer_signature
        if "%" not in encoded_signature:
            encoded_signature = urllib.parse.quote(encoded_signature)

        username_string = _add_to_username_parameter(
            username_string, encoded_signature, "x-amz-customauthorizer-signature=")

    if auth_token_key_name is not None and auth_token_value is not None:
        username_string = _add_to_username_parameter(username_string, auth_token_value, auth_token_key_name + "=")

    kwargs["username"] = username_string
    kwargs["password"] = auth_password

    tls_ctx_options = awscrt.io.TlsContextOptions()
    if use_websockets == False:
        kwargs["port"] = 443
        tls_ctx_options.alpn_list = ["mqtt"]

    def _sign_websocket_handshake_request(transform_args, **kwargs):
        # transform_args need to know when transform is done
        try:
            transform_args.set_done()
        except Exception as e:
            transform_args.set_done(e)

    return _builder(tls_ctx_options=tls_ctx_options,
                    use_websockets=use_websockets,
                    use_custom_authorizer=True,
                    websocket_handshake_transform=_sign_websocket_handshake_request if use_websockets else None,
                    **kwargs)


def new_default_builder(**kwargs) -> awscrt.mqtt.Connection:
    """
    This builder creates an :class:`awscrt.mqtt.Connection`, without any configuration besides the default TLS context options.

    This requires setting the connection details manually by passing all the necessary data
    in :mod:`common arguments<awsiot.mqtt_connection_builder>` to make a connection.
    """
    _check_required_kwargs(**kwargs)
    tls_ctx_options = awscrt.io.TlsContextOptions()

    return _builder(tls_ctx_options=tls_ctx_options,
                    use_websockets=False,
                    **kwargs)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0.
import awsiot
import boto3
import botocore.exceptions
import os.path
import shutil
import subprocess
import sys
import tempfile
import unittest
import uuid
import warnings


class Config:
    cache = None

    def __init__(self, endpoint, cert, key, region):
        self.endpoint = endpoint
        self.region = region
        self.cert_bytes = cert
        self.key_bytes = key

        self._tmp_dirpath = tempfile.mkdtemp()
        self.cert_filepath = os.path.join(self._tmp_dirpath, 'certificate.pem')
        with open(self.cert_filepath, 'wb') as cert_file:
            cert_file.write(cert)

        self.key_filepath = os.path.join(self._tmp_dirpath, 'privatekey.pem')
        with open(self.key_filepath, 'wb') as key_file:
            key_file.write(key)

    def __del__(self):
        shutil.rmtree(self._tmp_dirpath)

    @staticmethod
    def get():
        """Raises SkipTest if credentials aren't set up correctly"""
        if Config.cache:
            return Config.cache

        # boto3 caches the HTTPS connection for the API calls, which appears to the unit test
        # framework as a leak, so ignore it, that's not what we're testing here
        warnings.simplefilter('ignore', ResourceWarning)

        try:
            secrets = boto3.client('secretsmanager')
            response = secrets.get_secret_value(SecretId='unit-test/endpoint')
            endpoint = response['SecretString']
            response = secrets.get_secret_value(SecretId='unit-test/certificate')
            cert = response['SecretString'].encode('utf8')
            response = secrets.get_secret_value(SecretId='unit-test/privatekey')
            key = response['SecretString'].encode('utf8')
            region = secrets.meta.region_name
            Config.cache = Config(endpoint, cert, key, region)
        except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as ex:
            raise unittest.SkipTest("No credentials")

        return Config.cache


def create_client_id():
    return 'aws-crt-python-unit-test-{0}'.format(uuid.uuid4())


class SamplesTest(unittest.TestCase):

    def _run(self, args, stdout_checker):
        process = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()
        stdout = stdout.decode()
        try:
            self.assertEqual(0, process.returncode)
            stdout_checker(stdout)
        except Exception as e:
            # print output and rethrow exception
            print(subprocess.list2cmdline(args))
            print("--- stdout ---")
            for line in stdout.splitlines():
                print(line)
            print("--- stderr ---")
            for line in stderr.splitlines():
                print(line)
            print("--- end ---")

            raise e

    def test_pubsub(self):
        config = Config.get()
        args = [
            sys.executable,
            "samples/pubsub.py",
            "--endpoint", config.endpoint,
            "--cert", config.cert_filepath,
            "--key", config.key_filepath,
            "--client_id", create_client_id(),
            "--count", "1",
            "--verbosity", "Trace",
        ]

        def stdout_checker(stdout):
            # check for last line printed by sample
            last_line = stdout.splitlines()[-1]
            self.assertTrue(last_line.startswith("Disconnected!"))

        self._run(args, stdout_checker)

    def test_basic_discovery_response_only(self):
        config = Config.get()
        args = [
            sys.executable,
            "samples/basic_discovery.py",
            "--print_discover_resp_only",
            "--region", config.region,
            "--cert", config.cert_filepath,
            "--key", config.key_filepath,
            "--thing_name", "aws-sdk-crt-unit-test",
            "--verbosity", "Trace",
        ]

        def stdout_checker(stdout):
            # check for last line printed by sample
            last_line = stdout.splitlines()[-1]
            self.assertTrue(last_line.startswith("awsiot.greengrass_discovery.DiscoverResponse("))

        self._run(args, stdout_checker)
